from copy import copy
from enum import Flag, auto
import logging
import os
import sys
from typing import cast, Dict, List, Optional, Tuple, overload
# TODO: import from typing module when dropping support for python 3.7
from typing_extensions import Literal
import warnings

from ymmsl import (Identifier, Operator, SettingValue, Port, Reference,
                   Settings)

from libmuscle.api_guard import APIGuard
from libmuscle.checkpoint_triggers import TriggerManager
from libmuscle.communicator import Communicator, Message
from libmuscle.settings_manager import SettingsManager
from libmuscle.logging import LogLevel
from libmuscle.logging_handler import MuscleManagerHandler
from libmuscle.mpp_message import ClosePort
from libmuscle.mmp_client import MMPClient
from libmuscle.profiler import Profiler
from libmuscle.profiling import ProfileEventType
from libmuscle.snapshot_manager import SnapshotManager
from libmuscle.util import extract_log_file_location


_logger = logging.getLogger(__name__)


_FInitCacheType = Dict[Tuple[str, Optional[int]], Message]


class InstanceFlags(Flag):
    """Enumeration of properties that an instance may have.

    You may combine multiple flags using the bitwise OR operator `|`. For
    example:

    .. code-block:: python

        from libmuscle import (
                Instance, USES_CHECKPOINT_API, DONT_APPLY_OVERLAY)

        ports = ...
        flags = USES_CHECKPOINT_API | DONT_APPLY_OVERLAY
        instance = Instance(ports, flags)
    """

    DONT_APPLY_OVERLAY = auto()
    """Do not apply the received settings overlay during prereceive of F_INIT
    messages. If you're going to use :meth:`Instance.receive_with_settings` on
    your F_INIT ports, you need to set this flag when creating an
    :class:`Instance`.

    If you don't know what that means, do not specify this flag and everything
    will be fine. If it turns out that you did need to specify the flag, MUSCLE3
    will tell you about it in an error message and you can add it still.
    """

    USES_CHECKPOINT_API = auto()
    """Indicate that this instance supports checkpointing.

    You may not use any checkpointing API calls when this flag is not supplied.
    """

    KEEPS_NO_STATE_FOR_NEXT_USE = auto()
    """Indicate this instance does not carry state between iterations of the
    reuse loop.

    This corresponds to :external:py:attr:`ymmsl.KeepsStateForNextUse.NO`.

    If neither :attr:`KEEPS_NO_STATE_FOR_NEXT_USE` and
    :attr:`STATE_NOT_REQUIRED_FOR_NEXT_USE` are supplied, this corresponds to
    :external:py:attr:`ymmsl.KeepsStateForNextUse.REQUIRED`.
    """

    STATE_NOT_REQUIRED_FOR_NEXT_USE = auto()
    """Indicate this instance carries state between iterations of the
    reuse loop, however this state is not required for restarting.

    This corresponds to :external:py:attr:`ymmsl.KeepsStateForNextUse.HELPFUL`.

    If neither :attr:`KEEPS_NO_STATE_FOR_NEXT_USE` and
    :attr:`STATE_NOT_REQUIRED_FOR_NEXT_USE` are supplied, this corresponds to
    :external:py:attr:`ymmsl.KeepsStateForNextUse.REQUIRED`.
    """


_CHECKPOINT_SUPPORT_MASK = (
        InstanceFlags.USES_CHECKPOINT_API |
        InstanceFlags.KEEPS_NO_STATE_FOR_NEXT_USE |
        InstanceFlags.STATE_NOT_REQUIRED_FOR_NEXT_USE)


class Instance:
    """Represents a component instance in a MUSCLE3 simulation.

    This class provides a low-level send/receive API for the instance
    to use.
    """
    def __init__(
            self, ports: Optional[Dict[Operator, List[str]]] = None,
            flags: InstanceFlags = InstanceFlags(0)) -> None:
        """Create an Instance.

        Args:
            ports: A list of port names for each
                :external:py:class:`~ymmsl.Operator` of this component.
            flags: Indicate properties for this instance. See
                :py:class:`InstanceFlags` for a detailed description of possible
                flags.
        """
        self.__is_shut_down = False

        self._flags = InstanceFlags(flags)
        if InstanceFlags.USES_CHECKPOINT_API in self._flags:
            warnings.warn(
                    'Checkpointing in MUSCLE3 version 0.6.0 is still in'
                    ' development: the API may change in a future MUSCLE3'
                    ' release.')

        # Note that these are accessed by Muscle3, but otherwise private.
        self._name, self._index = self.__make_full_name()
        """Name and index of this instance."""

        mmp_location = self.__extract_manager_location()
        self.__manager = MMPClient(mmp_location)
        """Client object for talking to the manager."""

        self.__set_up_logging()

        self._api_guard = APIGuard(
                InstanceFlags.USES_CHECKPOINT_API in self._flags)
        """Checks that the user uses the API correctly."""

        self._profiler = Profiler(self._instance_name(), self.__manager)
        """Profiler for this instance."""

        self._communicator = Communicator(
                self._name, self._index, ports, self._profiler)
        """Communicator for this instance."""

        self._declared_ports = ports
        """Declared ports for this instance."""

        self._settings_manager = SettingsManager()
        """Settings for this instance."""

        self._snapshot_manager = SnapshotManager(
                self._instance_name(), self.__manager, self._communicator)
        """Resumes, loads and saves snapshots."""

        self._trigger_manager = TriggerManager()
        """Keeps track of checkpoints and triggers snapshots."""

        self._first_run = None          # type: Optional[bool]
        """Whether this is the first iteration of the reuse loop"""

        self._do_reuse = None           # type: Optional[bool]
        """Whether to enter this iteration of the reuse loop

        This is None during the reuse loop, and set between
        should_save_final_snapshot and reuse_instance.
        """

        self._do_resume = False
        """Whether to resume on this iteration of the reuse loop"""

        self._do_init = False
        """Whether to do f_init on this iteration of the reuse loop"""

        self._f_init_cache = dict()     # type: _FInitCacheType
        """Stores pre-received messages for f_init ports"""

        self._register()
        self._connect()

        # Note: get_checkpoint_info needs to have the ports initialized
        # so it comes after self._connect()
        checkpoint_info = self.__manager.get_checkpoint_info(
                self._instance_name())

        elapsed_time, checkpoints = checkpoint_info[0:2]
        self._trigger_manager.set_checkpoint_info(elapsed_time, checkpoints)

        if checkpoints and not (self._flags & _CHECKPOINT_SUPPORT_MASK):
            raise RuntimeError(
                    'The workflow has requested checkpoints, but this instance'
                    ' does not support checkpointing. Please consult the'
                    ' MUSCLE3 checkpointing documentation how to add'
                    ' checkpointing support.')

        resume_snapshot, snapshot_dir = checkpoint_info[2:4]
        saved_at = self._snapshot_manager.prepare_resume(
                resume_snapshot, snapshot_dir)
        # Resume settings overlay
        overlay = self._snapshot_manager._resume_overlay
        if overlay is not None:
            self._settings_manager.overlay = overlay

        if saved_at is not None:
            self._trigger_manager.update_checkpoints(saved_at)

        self._set_local_log_level()
        self._set_remote_log_level()

    def reuse_instance(self, apply_overlay: Optional[bool] = None) -> bool:
        """Decide whether to run this instance again.

        In a multiscale simulation, instances get reused all the time.
        For example, in a macro-micro simulation, the micromodel does a
        complete run for every timestep of the macromodel. Rather than
        starting up a new instance of the micromodel, which could be
        expensive, we reuse a single instance many times.

        This may bring other advantages, such as faster convergence
        when starting from the previous final state, and in some
        cases may be necessary if micromodel state needs to be
        preserved from one macro timestep to the next.

        So in MUSCLE, submodels run in a *reuse loop*, which runs them
        over and over again until their work is done and they should be
        shut down. Whether to do another F_INIT, O_I, S, O_F cycle
        is decided by this method.

        This method must be called at the beginning of the reuse loop,
        i.e. before the F_INIT operator, and its return value should
        decide whether to enter that loop again.

        Raises:
            RuntimeError:
                When implementing the checkpointing API, but libmuscle detected
                incorrect API calls. The description of the RuntimeError
                indicates which calls are incorrect or missing. For more
                information see the checkpointing API documentation in
                :meth:`resuming`, :meth:`load_snapshot`,
                :meth:`should_save_snapshot`, :meth:`save_snapshot`,
                :meth:`should_save_final_snapshot` and
                :meth:`save_final_snapshot`, or the checkpointing tutorial.
        """
        self._api_guard.verify_reuse_instance()

        if self._do_reuse is not None:
            # thank you, should_save_final_snapshot, for running this already
            do_reuse = self._do_reuse
            self._do_reuse = None
        else:
            do_reuse = self._decide_reuse_instance(apply_overlay)

        # now _first_run, _do_resume and _do_init are also set correctly

        do_implicit_checkpoint = (
                not self._first_run and
                InstanceFlags.USES_CHECKPOINT_API not in self._flags and
                (InstanceFlags.STATE_NOT_REQUIRED_FOR_NEXT_USE in self._flags or
                 InstanceFlags.KEEPS_NO_STATE_FOR_NEXT_USE in self._flags))

        if do_implicit_checkpoint:
            if self._trigger_manager.should_save_final_snapshot(
                    do_reuse, self.__f_init_max_timestamp):
                # store a None instead of a Message
                self._save_snapshot(None, True, self.__f_init_max_timestamp)

        if not do_reuse:
            self.__close_ports()
            self._communicator.shutdown()
            self._deregister()
            self.__manager.close()

        self._api_guard.reuse_instance_done(do_reuse)
        return do_reuse

    def error_shutdown(self, message: str) -> None:
        """Logs an error and shuts down the Instance.

        If you detect that something is wrong (invalid input, invalid
        settings, simulation diverged, or anything else really), you
        should call this method before calling exit() or raising
        an exception that you don't expect to catch.

        If you do so, the Instance will tell the rest of the simulation
        that it encountered an error and will shut down. That makes it
        easier to debug the situation (the message will be logged), and
        it reduces the chance that other parts of the simulation will
        sit around waiting forever for a message that this instance was
        supposed to send.

        Args:
            message: An error message describing the problem.
        """
        self.__shutdown(message)

    @overload
    def get_setting(self, name: str, typ: Literal['str']) -> str:
        ...

    @overload
    def get_setting(self, name: str, typ: Literal['int']) -> int:
        ...

    @overload
    def get_setting(self, name: str, typ: Literal['float']) -> float:
        ...

    @overload
    def get_setting(self, name: str, typ: Literal['bool']) -> bool:
        ...

    @overload
    def get_setting(self, name: str, typ: Literal['[float]']) -> List[float]:
        ...

    @overload
    def get_setting(
            self, name: str, typ: Literal['[[float]]']) -> List[List[float]]:
        ...

    @overload
    def get_setting(self, name: str, typ: None = None) -> SettingValue:
        ...

    def get_setting(self, name: str, typ: Optional[str] = None
                    ) -> SettingValue:
        """Returns the value of a model setting.

        Args:
            name: The name of the setting, without any instance
                    prefix.
            typ: The expected type of the value. If the value does
                    not match this type, a TypeError will be raised.
                    If not specified, any of the supported types
                    will be accepted, and you'll have to figure out
                    what you got yourself.

        Raises:
            KeyError: If no value was set for this setting.
            TypeError: If the type of the setting's value was not
                    as expected.
        """
        return self._settings_manager.get_setting(
                self._instance_name(), Reference(name), typ)

    def list_ports(self) -> Dict[Operator, List[str]]:
        """Returns a description of the ports that this Instance has.

        Note that the result has almost the same format as the port
        declarations you pass when making an Instance. The only
        difference is that the port names never have `[]` at the end,
        even if the port is a vector port.

        Returns:
            A dictionary, indexed by :external:py:class:`~ymmsl.Operator`,
            containing lists of port names. Operators with no associated ports
            are not included.
        """
        return self._communicator.list_ports()

    def is_connected(self, port: str) -> bool:
        """Returns whether the given port is connected.

        Args:
            port: The name of the port to inspect.

        Returns:
            True if there is a conduit attached to this port, False
            if not.
        """
        return self._communicator.get_port(port).is_connected()

    def is_vector_port(self, port: str) -> bool:
        """Returns whether a port is a vector or scalar port

        If a port has been declared to be a vector port (i.e. the
        name passed when creating this Instance had '[]' at the
        end), then you can pass a 'slot' argument when sending or
        receiving. It's like the port is a vector of slots on which
        you can send or receive messages.

        This function returns True if the given port is a vector
        port, and False if it is a scalar port.

        Args:
            port: The port to check this property of.
        """
        return self._communicator.get_port(port).is_vector()

    def is_resizable(self, port: str) -> bool:
        """Returns whether the given port is resizable.

        Scalar ports are never resizable. Whether a vector port is
        resizable depends on what it is connected to.

        Args:
            port: Name of the port to inspect.

        Returns:
            True if the port can be resized.
        """
        return self._communicator.get_port(port).is_resizable()

    def get_port_length(self, port: str) -> int:
        """Returns the current length of the port.

        Args:
            port: The name of the port to measure.

        Raises:
            RuntimeError: If this is a scalar port.
        """
        return self._communicator.get_port(port).get_length()

    def set_port_length(self, port: str, length: int) -> None:
        """Resizes the port to the given length.

        You should check whether the port is resizable using
        :meth:`is_resizable()` first; whether it is depends on how this
        component is wired up, so you should check.

        Args:
            port: Name of the port to resize.
            length: The new length.

        Raises:
            RuntimeError: If the port is not resizable.
        """
        self._communicator.get_port(port).set_length(length)

    def send(self, port_name: str, message: Message,
             slot: Optional[int] = None) -> None:
        """Send a message to the outside world.

        Sending is non-blocking, a copy of the message will be made
        and stored in memory until the receiver is ready to receive it.

        Args:
            port_name: The port on which this message is to be sent.
            message: The message to be sent.
            slot: The slot to send the message on, if any.
        """
        self.__check_port(port_name)
        if message.settings is None:
            message = copy(message)
            message.settings = self._settings_manager.overlay

        self._communicator.send_message(
                port_name, message, slot,
                self._trigger_manager.checkpoints_considered_until())

    def receive(self, port_name: str, slot: Optional[int] = None,
                default: Optional[Message] = None
                ) -> Message:
        """Receive a message from the outside world.

        Receiving is a blocking operation. This function will contact
        the sender, wait for a message to be available, and receive and
        return it.

        If the port you are receiving on is not connected, the default
        value you specified will be returned exactly as you passed it.
        If you didn't specify a default value (e.g. because there is no
        reasonable default, you really need the outside input) and the
        port is not connected, you'll get a RuntimeError.

        Args:
            port_name: The endpoint on which a message is to be
                    received.
            slot: The slot to receive the message on, if any.
            default: A default value to return if this port is not
                    connected.

        Returns:
            The received message. The settings attribute of the
            received message will be None.

        Raises:
            RuntimeError: If the given port is not connected and no
                    default value was given.
        """
        return self.__receive_message(port_name, slot, default, False)

    def receive_with_settings(
            self, port_name: str, slot: Optional[int] = None,
            default: Optional[Message] = None
            ) -> Message:
        """Receive a message with attached settings overlay.

        This function should not be used in submodels. It is intended
        for use by special component that are ensemble-aware and
        have to pass on overlay settings explicitly.

        Receiving is a blocking operation. This function will contact
        the sender, wait for a message to be available, and receive and
        return it.

        If the port you are receiving on is not connected, the default
        value you specified will be returned exactly as you passed it.
        If you didn't specify a default value (e.g. because there is no
        reasonable default, and you really need the outside input) and
        the port is not connected, then you'll get a RuntimeError.

        Args:
            port_name: The endpoint on which a message is to be
                    received.
            slot: The slot to receive the message on, if any.
            default: A default value to return if this port is not
                    connected.

        Returns:
            The received message. The settings attribute will
            contain the received Settings, and will not be None.

        Raises:
            RuntimeError: If the given port is not connected and no
                    default value was given.
        """
        return self.__receive_message(port_name, slot, default, True)

    def resuming(self) -> bool:
        """Check if this instance is resuming from a snapshot.

        Must be used by submodels that implement the checkpointing API. You'll
        get a RuntimeError when not calling this method in an iteration of the
        reuse loop.

        This method returns True for the first iteration of the reuse loop after
        resuming from a previously taken snapshot. When resuming from a
        snapshot, the submodel must load its state from the snapshot as returned
        by :meth:`load_snapshot`.

        Returns:
            True iff the submodel must resume from a snapshot instead of the
            usual F_INIT step during this iteration of the reuse loop.
        """
        self._api_guard.verify_resuming()
        self._api_guard.resuming_done(self._do_resume)
        return self._do_resume

    def should_init(self) -> bool:
        """Check if this instance should initialize.

        Must be used by submodels that implement the checkpointing API.

        When resuming from a previous snapshot, instances need not always
        execute the F_INIT phase of the submodel execution loop. Use this method
        before attempting to receive data on F_INIT ports.

        Returns:
            True if the submodel must execute the F_INIT step, False otherwise.
        """
        self._api_guard.verify_should_init()
        self._api_guard.should_init_done()
        return self._do_init

    def load_snapshot(self) -> Message:
        """Load a snapshot.

        Must only be called when :meth:`resuming` returns True.

        Returns:
            Message object containing the state as saved in a previous run
            through :meth:`save_snapshot` or :meth:`save_final_snapshot`

        Raises:
            RuntimeError: if not resuming from a snapshot.
        """
        self._api_guard.verify_load_snapshot()
        result = self._snapshot_manager.load_snapshot()
        self._api_guard.load_snapshot_done()
        return result

    def should_save_snapshot(self, timestamp: float) -> bool:
        """Check if a snapshot should be saved after the S Operator of the
        submodel.

        This method checks if a snapshot should be saved right now, based on the
        provided timestamp and passed wallclock time.

        When this method returns True, the submodel must also save a snapshot
        through :meth:`save_snapshot`. A RuntimeError will be generated when not
        doing so.

        See also :meth:`should_save_final_snapshot` for the variant that must be
        called at the end of the reuse loop.

        Args:
            timestamp: current timestamp of the submodel

        Returns:
            True iff a snapshot should be taken by the submodel according to the
            checkpoint rules provided in the ymmsl configuration.
        """
        self._api_guard.verify_should_save_snapshot()
        result = self._trigger_manager.should_save_snapshot(timestamp)
        self._api_guard.should_save_snapshot_done(result)
        return result

    def save_snapshot(self, message: Message) -> None:
        """Save a snapshot after the S Operator of the submodel.

        Before saving a snapshot, you should check using
        :meth:`should_save_snapshot` if a snapshot should be saved according to
        the checkpoint rules specified in the ymmsl configuration. You should
        use the same timestamp in the provided Message object as used to query
        `should_save_snapshot`.

        Although it is allowed to save a snapshot even when
        :meth:`should_save_snapshot` returns False, you should avoid this: this
        situation is not likely to lead to a consistent snapshot over all
        submodels of the run (and therefore it is not useful to restart from).
        It could also lead to a lot of snapshot files clogging your file system.

        See also :meth:`save_final_snapshot` for the variant that must be called
        at the end of the reuse loop.

        Args:
            message: Message object that is saved as snapshot. The message
                timestamp attribute should be the same as passed to
                :meth:`should_save_snapshot`. The data attribute can be used to
                store the internal state of the submodel.
        """
        self._api_guard.verify_save_snapshot()
        if message is None:
            raise RuntimeError('Please specify a Message to save as snapshot.')
        self._save_snapshot(message, False)
        self._api_guard.save_snapshot_done()

    def should_save_final_snapshot(self) -> bool:
        """Check if a snapshot should be saved at the end of the reuse loop.

        This method checks if a snapshot should be saved now.

        When this method returns True, the submodel must also save a snapshot
        through :meth:`save_final_snapshot`. A :class:`RuntimeError` will be
        generated when not doing so.

        See also :meth:`should_save_snapshot` for the variant that may be called
        inside of a time-integration loop of the submodel.

        .. note::
            This method will block until it can determine whether a final
            snapshot should be taken. This means it must also determine if this
            instance is reused. The optional keyword-only argument
            `apply_overlay` has the same meaning as for :meth:`reuse_instance`.

        Args:
            apply_overlay: Whether to apply the received settings
                overlay or to save it. If you're going to use
                :meth:`receive_with_settings` on your F_INIT ports, set this to
                False. If you don't know what that means, just call
                :meth:`should_save_final_snapshot()` without specifying this and
                everything will be fine. If it turns out that you did need to
                specify False, MUSCLE3 will tell you about it in an error
                message and you can add it still.

        Returns:
            True iff a final snapshot should be taken by the submodel according
            to the checkpoint rules provided in the ymmsl configuration.
        """
        self._api_guard.verify_should_save_final_snapshot()

        self._do_reuse = self._decide_reuse_instance()
        result = self._trigger_manager.should_save_final_snapshot(
                self._do_reuse, self.__f_init_max_timestamp)

        self._api_guard.should_save_final_snapshot_done(result)
        return result

    def save_final_snapshot(self, message: Message) -> None:
        """Save a snapshot at the end of the reuse loop.

        Before saving a snapshot, you should check using
        :meth:`should_save_final_snapshot` if a snapshot should be saved
        according to the checkpoint rules specified in the ymmsl configuration.

        Although it is allowed to save a snapshot even when
        :meth:`should_save_final_snapshot` returns False, you should avoid this:
        this situation is not likely to lead to a consistent snapshot over all
        submodels of the run (and therefore it is not useful to restart from).
        It could also lead to a lot of snapshot files clogging your file system.

        See also :meth:`save_snapshot` for the variant that may be called after
        each S Operator of the submodel.

        Args:
            message: Message object that is saved as snapshot. The data
                attribute can be used to store the internal state of the
                submodel.
        """
        self._api_guard.verify_save_final_snapshot()
        if message is None:
            raise RuntimeError('Please specify a Message to save as snapshot.')
        self._save_snapshot(message, True, self.__f_init_max_timestamp)
        self._api_guard.save_final_snapshot_done()

    @property
    def __f_init_max_timestamp(self) -> Optional[float]:
        """Return max timestamp of pre-received F_INIT messages
        """
        return max(
                (msg.timestamp for msg in self._f_init_cache.values()),
                default=None)

    def _register(self) -> None:
        """Register this instance with the manager.
        """
        register_event = self._profiler.start(ProfileEventType.REGISTER)
        locations = self._communicator.get_locations()
        port_list = self.__list_declared_ports()
        self.__manager.register_instance(
                self._instance_name(), locations, port_list)
        register_event.stop()
        _logger.info('Registered with the manager')

    def _connect(self) -> None:
        """Connect this instance to the given peers / conduits.
        """
        connect_event = self._profiler.start(ProfileEventType.CONNECT)
        conduits, peer_dims, peer_locations = self.__manager.request_peers(
                self._instance_name())
        self._communicator.connect(conduits, peer_dims, peer_locations)
        self._settings_manager.base = self.__manager.get_settings()
        connect_event.stop()
        _logger.info('Received peer locations and base settings')

    def _deregister(self) -> None:
        """Deregister this instance from the manager.
        """
        deregister_event = self._profiler.start(ProfileEventType.DEREGISTER)
        self.__manager.deregister_instance(self._instance_name())
        deregister_event.stop()
        # this is the last thing we'll profile, so flush messages
        self._profiler.shutdown()
        _logger.info('Deregistered from the manager')

    @staticmethod
    def __extract_manager_location() -> str:
        """Gets the manager network location from the command line.

        We use a --muscle-manager=<host:port> argument to tell the
        MUSCLE library how to connect to the manager. This function
        will extract this argument from the command line arguments,
        if it is present. If not, it will check the MUSCLE_MANAGER
        environment variable, and if that is not set, fall back to
        the default.

        Returns:
            A connection string, or None.
        """
        # Neither getopt, optparse, or argparse will let me pick out
        # just one option from the command line and ignore the rest.
        # So we do it by hand.
        prefix = '--muscle-manager='
        for arg in sys.argv[1:]:
            if arg.startswith(prefix):
                return arg[len(prefix):]

        return os.environ.get('MUSCLE_MANAGER', 'tcp:localhost:9000')

    def __set_up_logging(self) -> None:
        """Adds logging handlers for one or more instances.
        """
        id_str = str(self._instance_name())

        logfile = extract_log_file_location('muscle3.{}.log'.format(id_str))
        if logfile is not None:
            local_handler = logging.FileHandler(str(logfile), mode='w')
            formatter = logging.Formatter(
                    '%(asctime)-15s: %(levelname)-7s %(name)s: %(message)s')
            local_handler.setFormatter(formatter)
            logging.getLogger('libmuscle').addHandler(local_handler)
            logging.getLogger('ymmsl').addHandler(local_handler)

        if self.__manager is not None:
            self._mmp_handler = MuscleManagerHandler(id_str, logging.WARNING,
                                                     self.__manager)
            logging.getLogger().addHandler(self._mmp_handler)

    def _decide_reuse_instance(
            self, apply_overlay: Optional[bool] = None) -> bool:
        """Decide whether and how to reuse the instance.

        This sets self._first_run, self._do_resume and self._do_init, and
        returns whether to reuse one more time. This is the real top of
        the reuse loop, and it gets called by reuse_instance and
        should_save_final_snapshot.
        """
        if self._first_run is None:
            self._first_run = True
        elif self._first_run:
            self._first_run = False

        # resume from intermediate
        if self._first_run and self._snapshot_manager.resuming_from_intermediate():
            self._do_resume = True
            self._do_init = False
            return True

        f_init_connected = self._have_f_init_connections()

        # resume from final
        if self._first_run and self._snapshot_manager.resuming_from_final():
            if f_init_connected:
                got_f_init_messages = self._pre_receive(apply_overlay)
                self._do_resume = True
                self._do_init = True
                return got_f_init_messages
            else:
                self._do_resume = False     # unused
                self._do_init = False       # unused
                return False

        # fresh start or resuming from implicit snapshot
        self._do_resume = False

        # simple straight single run without resuming
        if not f_init_connected:
            self._do_init = self._first_run
            return self._first_run

        # not resuming and f_init connected, run while we get messages
        got_f_init_messages = self._pre_receive(apply_overlay)
        self._do_init = got_f_init_messages
        return got_f_init_messages

    def _save_snapshot(
            self, message: Optional[Message], final: bool,
            f_init_max_timestamp: Optional[float] = None) -> None:
        """Save a snapshot to disk and notify manager.

        Args:
            message: The data to save
            final: Whether this is a final snapshot or an intermediate
                one
            f_init_max_timestamp: Timestamp for final snapshots
        """
        triggers = self._trigger_manager.get_triggers()
        walltime = self._trigger_manager.elapsed_walltime()
        timestamp = self._snapshot_manager.save_snapshot(
                message, final, triggers, walltime,
                f_init_max_timestamp, self._settings_manager.overlay)
        self._trigger_manager.update_checkpoints(timestamp)

    def __receive_message(
            self, port_name: str, slot: Optional[int],
            default: Optional[Message], with_settings: bool
            ) -> Message:
        """Receives a message on the given port.

        This implements receive and receive_with_settings, see the
        description of those.
        """
        self.__check_port(port_name)

        port = self._communicator.get_port(port_name)
        if port.operator == Operator.F_INIT:
            if (port_name, slot) in self._f_init_cache:
                msg = self._f_init_cache[(port_name, slot)]
                del self._f_init_cache[(port_name, slot)]
                if with_settings and msg.settings is None:
                    err_msg = ('If you use receive_with_settings()'
                               ' on an F_INIT port, then you have to'
                               ' set the flag'
                               ' :attr:`InstanceFlag.DONT_APPLY_OVERLAY` when'
                               ' creating the :class:`Instance`, otherwise the'
                               ' settings will already have been applied by'
                               ' MUSCLE.')
                    self.__shutdown(err_msg)
                    raise RuntimeError(err_msg)
            else:
                if port.is_connected():
                    err_msg = (('Tried to receive twice on the same'
                                ' port "{}", that\'s not possible.'
                                ' Did you forget to call'
                                ' reuse_instance() in your reuse loop?'
                                ).format(port_name))
                    self.__shutdown(err_msg)
                    raise RuntimeError(err_msg)
                else:
                    if default is not None:
                        return default
                    err_msg = (('Tried to receive on port "{}",'
                                ' which is not connected, and no'
                                ' default value was given. Please'
                                ' connect this port!').format(port_name))
                    self.__shutdown(err_msg)
                    raise RuntimeError(err_msg)

        else:
            msg, saved_until = self._communicator.receive_message(
                    port_name, slot, default)
            if port.is_connected() and not port.is_open(slot):
                err_msg = (('Port {} was closed while trying to'
                            ' receive on it, did the peer crash?'
                            ).format(port_name))
                self.__shutdown(err_msg)
                raise RuntimeError(err_msg)
            if port.is_connected() and not with_settings:
                self.__check_compatibility(port_name, msg.settings)
            if not with_settings:
                msg.settings = None
            self._trigger_manager.harmonise_wall_time(saved_until)
        return msg

    def __make_full_name(self
                         ) -> Tuple[Reference, List[int]]:
        """Returns instance name and index.

        This takes the argument to the --muscle-instance= command-line
        option and splits it into a component name and an index.
        """
        def split_reference(ref: Reference) -> Tuple[Reference, List[int]]:
            index = list()     # type: List[int]
            i = 0
            while i < len(ref) and isinstance(ref[i], Identifier):
                i += 1
            name = ref[:i]

            while i < len(ref) and isinstance(ref[i], int):
                index.append(cast(int, ref[i]))
                i += 1

            return name, index

        # Neither getopt, optparse, or argparse will let me pick out
        # just one option from the command line and ignore the rest.
        # So we do it by hand.
        prefix_tag = '--muscle-instance='
        for arg in sys.argv[1:]:
            if arg.startswith(prefix_tag):
                prefix_str = arg[len(prefix_tag):]
                prefix_ref = Reference(prefix_str)
                name, index = split_reference(prefix_ref)
                break
        else:
            if 'MUSCLE_INSTANCE' in os.environ:
                prefix_ref = Reference(os.environ['MUSCLE_INSTANCE'])
                name, index = split_reference(prefix_ref)
            else:
                raise RuntimeError((
                    'A --muscle-instance command line argument or'
                    ' MUSCLE_INSTANCE environment variable is required to'
                    ' identify this instance. Please add one.'))
        return name, index

    def __list_declared_ports(self) -> List[Port]:
        """Returns a list of declared ports.

        This returns a list of ymmsl.Port objects, which have only the
        name and the operator, not libmuscle.Port, which has more.
        """
        result = list()
        if self._declared_ports is not None:
            for operator, port_names in self._declared_ports.items():
                for name in port_names:
                    if name.endswith('[]'):
                        name = name[:-2]
                    result.append(Port(Identifier(name), operator))
        return result

    def _instance_name(self) -> Reference:
        """Returns the full instance name.
        """
        return self._name + self._index

    def __check_port(self, port_name: str) -> None:
        if not self._communicator.port_exists(port_name):
            err_msg = (('Port "{}" does not exist on "{}". Please check'
                        ' the name and the list of ports you gave for'
                        ' this component.').format(port_name, self._name))
            self.__shutdown(err_msg)
            raise RuntimeError(err_msg)

    def _have_f_init_connections(self) -> bool:
        """Checks whether we have connected F_INIT ports.

        This includes muscle_settings_in, and any user-defined ports.
        """
        ports = self._communicator.list_ports()
        f_init_connected = any(
                [self.is_connected(port)
                 for port in ports.get(Operator.F_INIT, [])])
        return f_init_connected or self._communicator.settings_in_connected()

    def _pre_receive(self, apply_overlay: Optional[bool]) -> bool:
        """Pre-receives on all ports.

        This includes muscle_settings_in and all user-defined ports.

        Returns:
            True iff no ClosePort messages were received.
        """
        all_ports_open = self.__receive_settings()
        self.__pre_receive_f_init(apply_overlay)
        for message in self._f_init_cache.values():
            if isinstance(message.data, ClosePort):
                all_ports_open = False
        return all_ports_open

    def __receive_settings(self) -> bool:
        """Receives settings on muscle_settings_in.

        Returns:
            False iff the port is connnected and ClosePort was received.
        """
        default_message = Message(0.0, None, Settings(), Settings())
        message, saved_until = self._communicator.receive_message(
                'muscle_settings_in', None, default_message)
        if isinstance(message.data, ClosePort):
            return False
        if not isinstance(message.data, Settings):
            err_msg = ('"{}" received a message on'
                       ' muscle_settings_in that is not a'
                       ' Settings. It seems that your'
                       ' simulation is miswired or the sending'
                       ' instance is broken.'.format(self._instance_name()))
            self.__shutdown(err_msg)
            raise RuntimeError(err_msg)

        settings = cast(Settings, message.settings)
        for key, value in message.data.items():
            settings[key] = value
        self._settings_manager.overlay = settings

        self._trigger_manager.harmonise_wall_time(saved_until)
        return True

    def __pre_receive_f_init(self, apply_overlay: Optional[bool]) -> None:
        """Receives on all ports connected to F_INIT.

        This receives all incoming messages on F_INIT and stores them
        in self._f_init_cache.
        """
        if apply_overlay is not None:
            warnings.warn(
                    'Explicitly providing apply_overlay in reuse_instance is'
                    ' deprecated. Use InstanceFlags.DONT_APPLY_OVERLAY when'
                    ' creating the instance instead.', DeprecationWarning)
        else:
            apply_overlay = InstanceFlags.DONT_APPLY_OVERLAY not in self._flags

        def pre_receive(port_name: str, slot: Optional[int]) -> None:
            msg, saved_until = self._communicator.receive_message(
                    port_name, slot)
            self._f_init_cache[(port_name, slot)] = msg
            if apply_overlay:
                self.__apply_overlay(msg)
                self.__check_compatibility(port_name, msg.settings)
                msg.settings = None
            self._trigger_manager.harmonise_wall_time(saved_until)

        self._f_init_cache = dict()
        ports = self._communicator.list_ports()
        for port_name in ports.get(Operator.F_INIT, []):
            _logger.debug('Pre-receiving on port {}'.format(port_name))
            port = self._communicator.get_port(port_name)
            if not port.is_connected():
                continue
            if not port.is_vector():
                pre_receive(port_name, None)
            else:
                pre_receive(port_name, 0)
                # The above receives the length, if needed, so now we can
                # get the rest.
                for slot in range(1, port.get_length()):
                    pre_receive(port_name, slot)

    def _set_remote_log_level(self) -> None:
        """Sets the remote log level.

        This is the minimum level a message must have to be sent to
        the manager. It gets this from the muscle_remote_log_level
        setting.

        Note that this also sets the global log level to this level
        if it is currently higher, otherwise we still get no output.

        """
        try:
            log_level_str = self.get_setting('muscle_remote_log_level', 'str')
        except KeyError:
            # muscle_remote_log_level not set, do nothing and keep the default
            return

        try:
            log_level = LogLevel[log_level_str.upper()]
            if log_level == LogLevel.LOCAL:
                raise KeyError()

            py_level = log_level.as_python_level()
            self._mmp_handler.setLevel(py_level)
            if not logging.getLogger().isEnabledFor(py_level):
                logging.getLogger().setLevel(py_level)

        except KeyError:
            _logger.warning(
                ('muscle_remote_log_level is set to {}, which is not a'
                 ' valid remote log level. Please use one of DEBUG, INFO,'
                 ' WARNING, ERROR, CRITICAL, or DISABLED').format(
                     log_level_str))
            return

    def _set_local_log_level(self) -> None:
        """Sets the local log level.

        This sets the local log level for libmuscle and ymmsl, from the
        muscle_local_log_level setting.

        It also attaches a FileHandler which outputs to a local file named
        after the instance. This name can be overridden by the
        --muscle-log-file command line option.

        """
        try:
            log_level_str = self.get_setting('muscle_local_log_level', 'str')

            log_level = LogLevel[log_level_str.upper()]
            if log_level is None:
                _logger.warning(
                    ('muscle_remote_log_level is set to {}, which is not a'
                     ' valid log level. Please use one of DEBUG, INFO,'
                     ' WARNING, ERROR, CRITICAL, or DISABLED').format(
                         log_level_str))
                return

            py_level = log_level.as_python_level()
            logging.getLogger('libmuscle').setLevel(py_level)
            logging.getLogger('ymmsl').setLevel(py_level)
        except KeyError:
            # muscle_remote_log_level not set, do nothing and keep the default
            pass

    def __apply_overlay(self, message: Message) -> None:
        """Sets local overlay if we don't already have one.

        Args:
            message: The message to apply the overlay from.
        """
        if len(self._settings_manager.overlay) == 0:
            if message.settings is not None:
                self._settings_manager.overlay = message.settings

    def __check_compatibility(self, port_name: str,
                              overlay: Optional[Settings]) -> None:
        """Checks whether a received overlay matches the current one.

        Args:
            port_name: Name of the port on which the overlay was
                    received.
            overlay: The received overlay.
        """
        if overlay is None:
            return
        if self._settings_manager.overlay != overlay:
            err_msg = (('Unexpectedly received data from a'
                        ' parallel universe on port "{}". My'
                        ' settings are "{}" and I received'
                        ' from a universe with "{}".').format(
                            port_name, self._settings_manager.overlay,
                            overlay))
            self.__shutdown(err_msg)
            raise RuntimeError(err_msg)

    def __close_outgoing_ports(self) -> None:
        """Closes outgoing ports.

        This sends a close port message on all slots of all outgoing
        ports.
        """
        for operator, ports in self._communicator.list_ports().items():
            if operator.allows_sending():
                for port_name in ports:
                    port = self._communicator.get_port(port_name)
                    if port.is_vector():
                        for slot in range(port.get_length()):
                            self._communicator.close_port(port_name, slot)
                    else:
                        self._communicator.close_port(port_name)

    def __drain_incoming_port(self, port_name: str) -> None:
        """Receives messages until a ClosePort is received.

        Receives at least once.

        Args:
            port_name: Port to drain.
        """
        port = self._communicator.get_port(port_name)
        while port.is_open():
            # TODO: log warning if not a ClosePort
            self._communicator.receive_message(port_name)

    def __drain_incoming_vector_port(self, port_name: str) -> None:
        """Receives messages until a ClosePort is received.

        Works with (resizable) vector ports.

        Args:
            port_name: Port to drain.
        """
        port = self._communicator.get_port(port_name)
        while not all([not port.is_open(slot)
                       for slot in range(port.get_length())]):
            for slot in range(port.get_length()):
                if port.is_open(slot):
                    self._communicator.receive_message(port_name, slot)

    def __close_incoming_ports(self) -> None:
        """Closes incoming ports.

        This receives on all incoming ports until a ClosePort is
        received on them, signaling that there will be no more
        messages, and allowing the sending instance to shut down
        cleanly.
        """
        for operator, port_names in self._communicator.list_ports().items():
            if operator.allows_receiving():
                for port_name in port_names:
                    port = self._communicator.get_port(port_name)
                    if not port.is_connected():
                        continue
                    if not port.is_vector():
                        self.__drain_incoming_port(port_name)
                    else:
                        self.__drain_incoming_vector_port(port_name)

    def __close_ports(self) -> None:
        """Closes all ports.

        This sends a close port message on all slots of all outgoing
        ports, then receives one on all incoming ports.
        """
        self.__close_outgoing_ports()
        self.__close_incoming_ports()

    def __shutdown(self, message: str) -> None:
        """Shuts down simulation.

        This logs the given error message, communicates to the peers
        that we're shutting down, and deregisters from the manager.
        """
        if not self.__is_shut_down:
            _logger.critical(message)
            self.__close_ports()
            self._communicator.shutdown()
            self._deregister()
            self.__manager.close()
            self.__is_shut_down = True



from datetime import datetime
from pathlib import Path
import sys
from typing import Optional, Sequence

import click
import ymmsl
from ymmsl import Identifier, PartialConfiguration

from libmuscle.manager.manager import Manager
from libmuscle.manager.run_dir import RunDir


@click.command(no_args_is_help=True)
@click.argument(
        'ymmsl_files', nargs=-1, required=True, type=click.Path(
            exists=True, file_okay=True, dir_okay=False, readable=True,
            allow_dash=True, resolve_path=True))
@click.option(
        '--log-level', nargs=1, type=str, default='INFO', show_default=True,
        help='Set the local log level. Try DEBUG for (much) more output.')
@click.option(
        '--run-dir', nargs=1, type=click.Path(
            exists=True, file_okay=False, dir_okay=True, readable=True,
            writable=True, allow_dash=False, resolve_path=True),
        help=(
            'Directory to write logs, run metadata and output to. By default,'
            ' a directory with the name run_<model>_<date>-<time> will be'
            ' created.')
        )
@click.option(
        '--start-all/--no-start-all', default=False, help=(
            'Start all submodel instances listed in the configuration file(s).'
            )
        )
def manage_simulation(
        ymmsl_files: Sequence[str],
        start_all: bool,
        run_dir: Optional[str],
        log_level: Optional[str]
        ) -> None:
    """Run the MUSCLE3 Manager.

    The MUSCLE manager manages a coupled simulation. It can start the
    various submodels, and it helps them to find and connect to each
    other to exchange messages. The manager also distributes settings
    from the configuration file(s) to the submodel instances.

    The muscle_manager command takes a list of configuration files in
    yMMSL format (see the online documentation). The files are combined
    in the order given, so if for example a setting is given multiple
    times, then the value in the last file in the list to mention it is
    used.
    """
    configuration = PartialConfiguration()
    for path in ymmsl_files:
        with open(path, 'r') as f:
            configuration.update(ymmsl.load(f))

    if run_dir is None:
        if configuration.model is not None:
            model_name = configuration.model.name
        else:
            model_name = Identifier('model')
        timestamp = datetime.now()
        timestr = timestamp.strftime('%Y%m%d_%H%M%S')
        run_dir_path = Path.cwd() / 'run_{}_{}'.format(model_name, timestr)
    else:
        run_dir_path = Path(run_dir).resolve()

    run_dir_obj = RunDir(run_dir_path)
    if start_all:
        manager = Manager(configuration, run_dir_obj, log_level)
        manager.start_instances()
    else:
        if run_dir is None:
            manager = Manager(configuration, None, log_level)
        else:
            manager = Manager(configuration, run_dir_obj, log_level)
        print(manager.get_server_location())

    success = manager.wait()

    if not success:
        print('An error occurred during execution, and the simulation was')
        print('shut down. The manager log should tell you what happened.')
        print('You can find it at')
        print(run_dir_path / 'muscle3_manager.log')

    sys.exit(0 if success else 1)


if __name__ == '__main__':
    manage_simulation()



from queue import Queue


class Outbox:
    """Stores messages to be sent to a particular receiver.

    An Outbox is a queue of messages, which may be deposited and
    then retrieved in the same order.
    """
    def __init__(self) -> None:
        """Create an empty Outbox.
        """
        self.__queue = Queue()  # type: Queue[bytes]

    def is_empty(self) -> bool:
        """Returns True iff the outbox is empty.
        """
        return self.__queue.empty()

    def deposit(self, message: bytes) -> None:
        """Put a message in the Outbox.

        The message will be placed at the back of a queue, and may be
        retrieved later via :py:meth:`retrieve`.

        Args:
            message: The message to store.
        """
        self.__queue.put(message)

    def retrieve(self) -> bytes:
        """Retrieve a message from the Outbox.

        The message will be removed from the front of the queue, and
        returned to the caller. Blocks if the queue is empty, until a
        message is deposited.

        Returns:
            The next message.
        """
        return self.__queue.get()



from collections import OrderedDict

import numpy as np
from ymmsl import (Component, Conduit, Configuration, Model, Operator,
                   Settings)

from libmuscle import Grid, Instance, Message
from libmuscle.runner import run_simulation


def macro():
    """Macro model implementation.
    """
    instance = Instance({
            Operator.O_I: ['out[]'],
            Operator.S: ['in[]']})

    while instance.reuse_instance():
        # f_init
        assert instance.get_setting('test1') == 13

        # o_i
        assert instance.is_vector_port('out')
        for slot in range(10):
            instance.send('out', Message(0.0, 10.0, 'testing'), slot)

        # s/b
        for slot in range(10):
            msg = instance.receive('in', slot)
            assert msg.data['string'] == 'testing back'
            assert msg.data['int'] == 42
            assert msg.data['float'] == 3.1416
            assert msg.data['grid'].array.dtype == np.float64
            assert msg.data['grid'].array[0, 1] == 34.0


def micro():
    """Micro model implementation.
    """
    instance = Instance({
            Operator.F_INIT: ['in'],
            Operator.O_F: ['out']})

    while instance.reuse_instance():
        # f_init
        assert instance.get_setting('test3', 'str') == 'testing'
        assert instance.get_setting('test4', 'bool') is True
        assert instance.get_setting('test6', '[[float]]')[0][1] == 2.0

        msg = instance.receive('in')
        assert msg.data == 'testing'

        # o_f
        result = {
                'string': 'testing back',
                'int': 42,
                'float': 3.1416,
                'grid': Grid(np.array([[12.0, 34.0, 56.0], [1.0, 2.0, 3.0]]))}
        instance.send('out', Message(0.1, data=result))


def test_all(log_file_in_tmpdir):
    """A positive all-up test of everything.
    """
    elements = [
            Component('macro', 'macro_impl'),
            Component('micro', 'micro_impl', [10])]

    conduits = [
            Conduit('macro.out', 'micro.in'),
            Conduit('micro.out', 'macro.in')]

    model = Model('test_model', elements, conduits)
    settings = Settings(OrderedDict([
                ('test1', 13),
                ('test2', 13.3),
                ('test3', 'testing'),
                ('test4', True),
                ('test5', [2.3, 5.6]),
                ('test6', [[1.0, 2.0], [3.0, 1.0]])]))

    configuration = Configuration(model, settings)

    implementations = {'macro_impl': macro, 'micro_impl': micro}
    run_simulation(configuration, implementations)



from pathlib import Path

import pytest
from ymmsl import Component, Conduit, Configuration, Model, Reference

from libmuscle.manager.instance_registry import InstanceRegistry
from libmuscle.manager.logger import Logger
from libmuscle.manager.mmp_server import MMPRequestHandler
from libmuscle.manager.run_dir import RunDir
from libmuscle.manager.snapshot_registry import SnapshotRegistry
from libmuscle.manager.topology_store import TopologyStore


@pytest.fixture
def logger(tmpdir):
    test_logger = Logger(Path(str(tmpdir)))
    yield test_logger
    test_logger.close()


@pytest.fixture
def mmp_configuration():
    return Configuration(
            Model(
                'test_model',
                [
                    Component('macro', 'macro_implementation'),
                    Component(
                        'micro', 'micro_implementation', [10, 10])],
                [
                    Conduit('macro.out', 'micro.in'),
                    Conduit('micro.out', 'macro.in')
                ]))


@pytest.fixture
def instance_registry():
    return InstanceRegistry()


@pytest.fixture
def topology_store(mmp_configuration) -> TopologyStore:
    return TopologyStore(mmp_configuration)


@pytest.fixture
def snapshot_registry(mmp_configuration, topology_store) -> SnapshotRegistry:
    return SnapshotRegistry(mmp_configuration, None, topology_store)


@pytest.fixture
def mmp_request_handler(
        logger, mmp_configuration, instance_registry, topology_store,
        snapshot_registry):
    return MMPRequestHandler(
            logger, mmp_configuration, instance_registry, topology_store,
            snapshot_registry, None)


@pytest.fixture
def loaded_instance_registry(instance_registry):
    instance_registry.add(Reference('macro'), ['direct:macro'], [])
    for j in range(10):
        for i in range(10):
            name = Reference('micro') + j + i
            location = 'direct:{}'.format(name)
            instance_registry.add(name, [location], [])
    return instance_registry


@pytest.fixture
def registered_mmp_request_handler(
        logger, mmp_configuration, loaded_instance_registry, topology_store,
        snapshot_registry):
    return MMPRequestHandler(
            logger, mmp_configuration, loaded_instance_registry, topology_store,
            snapshot_registry, None)


@pytest.fixture
def mmp_configuration2():
    return Configuration(
            Model(
                'test_model',
                [
                    Component('macro', 'macro_implementation'),
                    Component('meso', 'meso_implementation', [5]),
                    Component('micro', 'micro_implementation', [5, 10])
                ],
                [
                    Conduit('macro.out', 'meso.in'),
                    Conduit('meso.out', 'micro.in'),
                    Conduit('micro.out', 'meso.in'),
                    Conduit('meso.out', 'macro.in')
                ]))


@pytest.fixture
def topology_store2(mmp_configuration2) -> TopologyStore:
    return TopologyStore(mmp_configuration2)


@pytest.fixture
def snapshot_registry2(mmp_configuration2, topology_store) -> SnapshotRegistry:
    return SnapshotRegistry(mmp_configuration2, None, topology_store)


@pytest.fixture
def loaded_instance_registry2():
    instance_registry = InstanceRegistry()

    instance_registry.add(Reference('macro'), ['direct:macro'], [])

    for j in range(5):
        name = Reference('meso') + j
        location = 'direct:{}'.format(name)
        instance_registry.add(name, [location], [])

    for j in range(5):
        for i in range(10):
            name = Reference('micro') + j + i
            location = 'direct:{}'.format(name)
            instance_registry.add(name, [location], [])
    return instance_registry


@pytest.fixture
def registered_mmp_request_handler2(
        logger, mmp_configuration, loaded_instance_registry2, topology_store2,
        snapshot_registry2, tmp_path):
    return MMPRequestHandler(
            logger, mmp_configuration,
            loaded_instance_registry2, topology_store2, snapshot_registry2,
            RunDir(tmp_path))



import logging
import multiprocessing as mp
import os
import subprocess
import sys
from contextlib import contextmanager, ExitStack
from pathlib import Path

import pytest
import yatiml
import ymmsl

import integration_test.include_libmuscle   # noqa: F401

from libmuscle.manager.manager import Manager
from libmuscle.manager.run_dir import RunDir


skip_if_python_only = pytest.mark.skipif(
        'MUSCLE_TEST_PYTHON_ONLY' in os.environ,
        reason='Python-only tests requested')


@pytest.fixture
def yatiml_log_warning():
    yatiml.logger.setLevel(logging.WARNING)


def ls_snapshots(run_dir, instance=None):
    """List all snapshots of the instance or workflow"""
    return sorted(run_dir.snapshot_dir(instance).iterdir(),
                  key=lambda path: tuple(map(int, path.stem.split("_")[1:])))


def start_mmp_server(control_pipe, ymmsl_doc, run_dir):
    control_pipe[0].close()
    manager = Manager(ymmsl_doc, run_dir)
    control_pipe[1].send(manager.get_server_location())
    control_pipe[1].recv()
    control_pipe[1].close()
    manager.stop()


def make_server_process(ymmsl_doc, tmpdir):
    run_dir = RunDir(Path(tmpdir))
    control_pipe = mp.Pipe()
    process = mp.Process(target=start_mmp_server,
                         args=(control_pipe, ymmsl_doc, run_dir),
                         name='Manager')
    process.start()
    control_pipe[1].close()
    # wait for start
    yield control_pipe[0].recv()
    control_pipe[0].send(True)
    control_pipe[0].close()
    process.join()


def _python_wrapper(instance_name, muscle_manager, callable):
    sys.argv.append(f'--muscle-instance={instance_name}')
    sys.argv.append(f'--muscle-manager={muscle_manager}')
    callable()


def run_manager_with_actors(
        ymmsl_text, tmpdir,
        cpp_actors={}, fortran_actors={}, python_actors={}):
    """Start muscle_manager along with C++ and python actors.

    C++ actors are a dict of instance->executable_path. Executable paths are
    assumed to be relative to ../libmuscle/cpp/build/. LD_LIBRARY_PATH is
    automatically updated to include the msgpack library path.

    Fortran actors are a dict of instance->executable_path. Executable paths are
    assumed to be relative to ../libmuscle/fortran/build/. LD_LIBRARY_PATH is
    automatically updated to include the msgpack library path.

    Python actors are a dict of instance->callable, where the callable
    implements the python actor.
    """
    env = os.environ.copy()
    ymmsl_doc = ymmsl.load(ymmsl_text)
    libmuscle_dir = Path(__file__).parents[1] / 'libmuscle'
    cpp_build_dir = libmuscle_dir / 'cpp' / 'build'
    fortran_build_dir = libmuscle_dir / 'fortran' / 'build'

    with ExitStack() as stack:
        # start muscle_manager and extract manager location
        ctx = contextmanager(make_server_process)(ymmsl_doc, tmpdir)
        env['MUSCLE_MANAGER'] = stack.enter_context(ctx)

        lib_paths = [cpp_build_dir / 'msgpack' / 'msgpack' / 'lib']
        if 'LD_LIBRARY_PATH' in env:
            env['LD_LIBRARY_PATH'] += ':' + ':'.join(map(str, lib_paths))
        else:
            env['LD_LIBRARY_PATH'] = ':'.join(map(str, lib_paths))

        native_processes = []
        # start native actors
        for actors, build_dir in ((cpp_actors, cpp_build_dir),
                                  (fortran_actors, fortran_build_dir)):
            for instance_name, executable_path in actors.items():
                executable = build_dir / executable_path
                f_out = stack.enter_context(
                        (tmpdir / f'{instance_name}_stdout.txt').open('w'))
                f_err = stack.enter_context(
                        (tmpdir / f'{instance_name}_stderr.txt').open('w'))
                native_processes.append(subprocess.Popen(
                        [str(executable), f'--muscle-instance={instance_name}'],
                        env=env, stdout=f_out, stderr=f_err))

        # start python actors
        python_processes = []
        for instance_name, callable in python_actors.items():
            proc = mp.Process(
                    target=_python_wrapper,
                    args=(instance_name, env['MUSCLE_MANAGER'], callable),
                    name=instance_name)
            proc.start()
            python_processes.append(proc)

        # check results
        for proc in native_processes:
            proc.wait()
            assert proc.returncode == 0
        for proc in python_processes:
            proc.join()
            assert proc.exitcode == 0


@pytest.fixture
def mmp_server_config(yatiml_log_warning):
    return (
            'ymmsl_version: v0.1\n'
            'model:\n'
            '  name: test_model\n'
            '  components:\n'
            '    macro: macro_implementation\n'
            '    micro:\n'
            '      implementation: micro_implementation\n'
            '      multiplicity: [10]\n'
            '  conduits:\n'
            '    macro.out: micro.in\n'
            '    micro.out: macro.in\n'
            'settings:\n'
            '  test1: 13\n'
            '  test2: 13.3\n'
            '  test3: testing\n'
            '  test4: True\n'
            '  test5: [2.3, 5.6]\n'
            '  test6:\n'
            '    - [1.0, 2.0]\n'
            '    - [3.0, 1.0]\n'
            'implementations:\n'
            '  macro_implementation: macro.py\n'
            '  micro_implementation: micro.py\n'
            )


@pytest.fixture
def mmp_server_process(mmp_server_config, tmpdir):
    ymmsl_doc = ymmsl.load(mmp_server_config)
    yield from make_server_process(ymmsl_doc, tmpdir)


@pytest.fixture
def mmp_server_config_simple(yatiml_log_warning):
    return (
            'ymmsl_version: v0.1\n'
            'model:\n'
            '  name: test_model\n'
            '  components:\n'
            '    macro: macro_implementation\n'
            '    micro: micro_implementation\n'
            '  conduits:\n'
            '    macro.out: micro.in\n'
            '    micro.out: macro.in\n'
            'settings:\n'
            '  test1: 13\n'
            '  test2: 13.3\n'
            '  test3: testing\n'
            '  test4: True\n'
            '  test5: [2.3, 5.6]\n'
            '  test6:\n'
            '    - [1.0, 2.0]\n'
            '    - [3.0, 1.0]\n'
            )


@pytest.fixture
def mmp_server_process_simple(mmp_server_config_simple, tmpdir):
    ymmsl_doc = ymmsl.load(mmp_server_config_simple)
    yield from make_server_process(ymmsl_doc, tmpdir)


@pytest.fixture
def mmp_server(mmp_server_config_simple, yatiml_log_warning):
    ymmsl_doc = ymmsl.load(mmp_server_config_simple)

    manager = Manager(ymmsl_doc)
    yield manager._server
    manager.stop()


@pytest.fixture
def log_file_in_tmpdir(tmpdir):
    old_workdir = os.getcwd()
    os.chdir(tmpdir)

    yield None

    os.chdir(old_workdir)


@pytest.fixture
def mpi_is_intel():
    if 'MUSCLE_ENABLE_CPP_MPI' not in os.environ:
        return None

    result = subprocess.run(
            ['mpirun', '--version'], capture_output=True, check=True)
    return 'Intel' in result.stdout.decode('utf-8')


@pytest.fixture
def mpirun_outfile_arg(mpi_is_intel):
    if mpi_is_intel:
        return '-outfile-pattern'
    else:
        return '--output-filename'


@pytest.fixture
def mpi_exec_model(mpi_is_intel):
    if mpi_is_intel:
        return 'intelmpi'
    else:
        return 'openmpi'



from threading import Condition
from typing import Dict  # noqa
from typing import List

from ymmsl import Port, Reference


class AlreadyRegistered(RuntimeError):
    pass


class InstanceRegistry:
    """Keeps track of running instances.

    The InstanceRegistry is a simple in-memory database that stores
    information about running instances of simulation components.
    """
    def __init__(self) -> None:
        """Construct an empty InstanceRegistry"""
        self._deregistered_one = Condition()    # doubles as lock
        self._locations = dict()  # type: Dict[Reference, List[str]]
        self._ports = dict()  # type: Dict[Reference, List[Port]]
        self._startup = True

    def add(self, name: Reference, locations: List[str], ports: List[Port]
            ) -> None:
        """Add an instance to the registry.

        Args:
            name: Name of the instance.
            locations: Network locations where it can be reached.
            ports: List of ports of this instance.

        Raises:
            ValueError: If an instance with this name has already been
                    registered.
        """
        with self._deregistered_one:
            if name in self._locations:
                raise AlreadyRegistered(
                        'Instance {} tried to register twice'.format(name))

            self._locations[name] = locations
            self._ports[name] = ports
            self._startup = False

    def get_locations(self, name: Reference) -> List[str]:
        """Retrieves the locations of a registered instance.

        Args:
            name: The name of the instance to get the location of.

        Raises:
            KeyError: If no instance with this name was registered.
        """
        with self._deregistered_one:
            return self._locations[name]

    def get_ports(self, name: Reference) -> List[Port]:
        """Retrieves the ports of a registered instance.

        Args:
            name: The name of the instance whose ports to retrieve.

        Raises:
            KeyError: If no instance with this name was registered.
        """
        with self._deregistered_one:
            return self._ports[name]

    def remove(self, name: Reference) -> None:
        """Remove an instance from the registry.

        Args:
            name: Name of the instance to remove.

        Raises:
            KeyError: If the instance does not exist.
        """
        with self._deregistered_one:
            del self._locations[name]
            del self._ports[name]
            self._deregistered_one.notify()

    def wait(self) -> None:
        """Waits until all instances are deregistered.

        This function blocks, and returns after each instance has been
        registered and deregistered again, signally the end of the
        simulation run.
        """
        # This is called from a different thread than add/remove
        with self._deregistered_one:
            while self._startup or self._locations:
                self._deregistered_one.wait()



import socket
import socketserver as ss
import threading
from typing import cast, List, Optional, Tuple
from typing_extensions import Type

import netifaces

from libmuscle.mcp.transport_server import RequestHandler, TransportServer
from libmuscle.mcp.tcp_util import (recv_all, recv_int64, send_int64,
                                    SocketClosed)


class TcpTransportServerImpl(ss.ThreadingMixIn, ss.TCPServer):
    daemon_threads = True
    allow_reuse_address = True

    def __init__(self, host_port_tuple: Tuple[str, int],
                 streamhandler: Type, transport_server: 'TcpTransportServer'
                 ) -> None:
        super().__init__(host_port_tuple, streamhandler)
        self.transport_server = transport_server
        if hasattr(socket, "TCP_NODELAY"):
            self.socket.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
        if hasattr(socket, "TCP_QUICKACK"):
            self.socket.setsockopt(socket.SOL_TCP, socket.TCP_QUICKACK, 1)


class TcpHandler(ss.BaseRequestHandler):
    """Handler for MCP-over-TCP connections.

    This is a Python handler for Python's TCPServer, which forwards
    to the RequestHandler attached to the server.
    """
    def handle(self) -> None:
        """Handles requests on a socket
        """
        request = self.receive_request()

        while request is not None:
            server = cast(TcpTransportServerImpl, self.server).transport_server
            response = server._handler.handle_request(request)

            send_int64(self.request, len(response))
            self.request.sendall(response)
            request = self.receive_request()

    def receive_request(self) -> Optional[bytes]:
        """Receives a request

        Returns:
            The received bytes
        """
        try:
            length = recv_int64(self.request)
            reqbuf = recv_all(self.request, length)
            return reqbuf
        except SocketClosed:
            return None


class TcpTransportServer(TransportServer):
    """A TransportServer that uses TCP to communicate."""
    def __init__(self, handler: RequestHandler, port: int = 0) -> None:
        """Create a TCPServer.

        Args:
            handler: A RequestHandler to handle requests
            port: The port to use.

        Raises:
            OSError: With errno set to errno.EADDRINUSE if the port is not
                available.
        """
        super().__init__(handler)

        self._server = TcpTransportServerImpl(('', port), TcpHandler, self)
        self._server_thread = threading.Thread(
                target=self._server.serve_forever, args=(0.1,), daemon=True)
        self._server_thread.start()

    def get_location(self) -> str:
        """Returns the location this server listens on.

        Returns:
            A string containing the location.
        """
        host, port = self._server.server_address

        locs = list()   # type: List[str]
        for address in self._get_if_addresses():
            locs.append('{}:{}'.format(address, port))
        return 'tcp:{}'.format(','.join(locs))

    def close(self) -> None:
        """Closes this server.

        Stops the server listening, waits for existing clients to
        disconnect, then frees any other resources.
        """
        self._server.shutdown()
        self._server_thread.join()
        self._server.server_close()

    def _get_if_addresses(self) -> List[str]:
        all_addresses = list()  # type: List[str]
        ifs = netifaces.interfaces()
        for interface in ifs:
            addrs = netifaces.ifaddresses(interface)
            for props in addrs.get(netifaces.AF_INET, []):
                if not props['addr'].startswith('127.'):
                    all_addresses.append(props['addr'])
            for props in addrs.get(netifaces.AF_INET6, []):
                # filter out link-local addresses with a scope id
                if '%' not in props['addr'] and props['addr'] != '::1':
                    all_addresses.append('[' + props['addr'] + ']')
        return all_addresses



import logging
from typing import Any, Dict, List, Optional, Tuple, cast
from ymmsl import Conduit, Identifier, Operator, Reference, Settings

from libmuscle.endpoint import Endpoint
from libmuscle.mpp_message import ClosePort, MPPMessage
from libmuscle.mpp_client import MPPClient
from libmuscle.mcp.transport_server import TransportServer
from libmuscle.mcp.type_registry import transport_server_types
from libmuscle.peer_manager import PeerManager
from libmuscle.post_office import PostOffice
from libmuscle.port import Port
from libmuscle.profiler import Profiler
from libmuscle.profiling import ProfileEventType


_logger = logging.getLogger(__name__)


MessageObject = Any


class Message:
    """A message to be sent or received.

    This class describes a message to be sent or that has been
    received.

    Attributes:
        timestamp (float): Simulation time for which this data is valid.
        next_timestamp (Optional[float]): Simulation time for the next
                message to be transmitted through this port.
        data (MessageObject): An object to send or that was received.
        settings (Settings): Overlay settings to send or that was
                received.
    """
    # Note: This is for communication with the user, it's not what
    # actually goes out on the wire, see libmuscle.mcp.Message for that.
    def __init__(self, timestamp: float, next_timestamp: Optional[float] = None,
                 data: MessageObject = None,
                 settings: Optional[Settings] = None
                 ) -> None:
        """Create a Message.

        Args:
            timestamp: Simulation time for which this data is valid.
            next_timestamp: Simulation time for the next message to be
                    transmitted through this port.
            data: An object to send or that was received.
            settings: Overlay settings to send or that were received.
        """
        # make sure timestamp and next_timestamp are floats
        timestamp = float(timestamp)
        if next_timestamp is not None:
            next_timestamp = float(next_timestamp)

        self.timestamp = timestamp
        self.next_timestamp = next_timestamp
        self.data = data
        self.settings = settings


class Communicator:
    """Communication engine for MUSCLE3.

    This class is the mailroom for a kernel that uses MUSCLE3. It
    manages the sending and receiving of messages, although it
    leaves the actual data transmission to various protocol-specific
    servers and clients.
    """
    def __init__(self, kernel: Reference, index: List[int],
                 declared_ports: Optional[Dict[Operator, List[str]]],
                 profiler: Profiler) -> None:
        """Create a Communicator.

        The instance reference must start with one or more Identifiers,
        giving the kernel id, followed by one or more integers which
        specify the instance index.

        Args:
            kernel: The kernel this is the Communicator for.
            index: The index for this instance.
            declared_ports: The declared ports for this instance
            profiler: The profiler to use for recording sends and
                    receives.
        """
        self._kernel = kernel
        self._index = index
        self._declared_ports = declared_ports
        self._post_office = PostOffice()
        self._profiler = profiler

        self._servers = list()  # type: List[TransportServer]

        # indexed by remote instance id
        self._clients = dict()  # type: Dict[Reference, MPPClient]

        for server_type in transport_server_types:
            server = server_type(self._post_office)
            self._servers.append(server)

        self._ports = dict()   # type: Dict[str, Port]

    def get_locations(self) -> List[str]:
        """Returns a list of locations that we can be reached at.

        These locations are of the form 'protocol:location', where
        the protocol name does not contain a colon and location may
        be an arbitrary string.

        Returns:
            A list of strings describing network locations.
        """
        return [server.get_location() for server in self._servers]

    def connect(self, conduits: List[Conduit],
                peer_dims: Dict[Reference, List[int]],
                peer_locations: Dict[Reference, List[str]]) -> None:
        """Connect this Communicator to its peers.

        This is the second stage in the simulation wiring process.

        Peers here are instances, and peer_dims and peer_locations are
        indexed by a Reference to an instance. Instance sets are
        multi-dimensional arrays with sizes given by peer_dims.

        Args:
            conduits: A list of conduits attached to this component,
                    as received from the manager.
            peer_dims: For each peer we share a conduit with, the
                    dimensions of the instance set.
            peer_locations: A list of locations for each peer instance
                    we share a conduit with.
        """
        self._peer_manager = PeerManager(
                self._kernel, self._index, conduits, peer_dims,
                peer_locations)

        if self._declared_ports is not None:
            self._ports = self.__ports_from_declared()
        else:
            self._ports = self.__ports_from_conduits(conduits)

        self._muscle_settings_in = self.__settings_in_port(conduits)

    def settings_in_connected(self) -> bool:
        """Returns True iff muscle_settings_in is connected.
        """
        return self._muscle_settings_in.is_connected()

    def list_ports(self) -> Dict[Operator, List[str]]:
        """Returns a description of the ports this Communicator has.

        Returns:
            A dictionary, indexed by Operator, containing lists of
            port names. Operators with no associated ports are not
            included.
        """
        result = dict()     # type: Dict[Operator, List[str]]
        for port_name, port in self._ports.items():
            if port.operator not in result:
                result[port.operator] = list()
            result[port.operator].append(port_name)
        return result

    def port_exists(self, port_name: str) -> bool:
        """Returns whether a port with the given name exists.

        Args:
            port_name: Port name to check.
        """
        return port_name in self._ports

    def get_port(self, port_name: str) -> Port:
        """Returns a Port object describing a port with the given name.

        Args:
            port: The port to retrieve.
        """
        return self._ports[port_name]

    def send_message(
            self, port_name: str, message: Message,
            slot: Optional[int] = None,
            checkpoints_considered_until: float = float('-inf')) -> None:
        """Send a message and settings to the outside world.

        Sending is non-blocking, a copy of the message will be made
        and stored until the receiver is ready to receive it.

        Args:
            port_name: The port on which this message is to be sent.
            message: The message to be sent.
            slot: The slot to send the message on, if any.
            checkpoints_considered_until: When we last checked if we
                should save a snapshot (wallclock time).
        """
        if slot is None:
            _logger.debug('Sending message on {}'.format(port_name))
            slot_list = []  # type: List[int]
        else:
            _logger.debug('Sending message on {}[{}]'.format(port_name, slot))
            slot_list = [slot]
            slot_length = self._ports[port_name].get_length()
            if slot_length <= slot:
                raise RuntimeError(('Slot out of bounds. You are sending on'
                                    ' slot {} of port "{}", which is of length'
                                    ' {}, so that slot does not exist'
                                    ).format(slot, port_name, slot_length))

        snd_endpoint = self.__get_endpoint(port_name, slot_list)
        if not self._peer_manager.is_connected(snd_endpoint.port):
            # log sending on disconnected port
            return

        port = self._ports[port_name]
        profile_event = self._profiler.start(ProfileEventType.SEND, port,
                                             None, slot, None)

        recv_endpoints = self._peer_manager.get_peer_endpoints(
                snd_endpoint.port, slot_list)

        port_length = None
        if port.is_resizable():
            port_length = port.get_length()

        for recv_endpoint in recv_endpoints:
            mcp_message = MPPMessage(snd_endpoint.ref(), recv_endpoint.ref(),
                                     port_length,
                                     message.timestamp, message.next_timestamp,
                                     cast(Settings, message.settings),
                                     port.get_num_messages(slot),
                                     checkpoints_considered_until,
                                     message.data)
            encoded_message = mcp_message.encoded()
            self._post_office.deposit(recv_endpoint.ref(), encoded_message)

        port.increment_num_messages(slot)

        profile_event.stop()
        if port.is_vector():
            profile_event.port_length = port.get_length()
        profile_event.message_size = len(encoded_message)

    def receive_message(self, port_name: str, slot: Optional[int] = None,
                        default: Optional[Message] = None
                        ) -> Tuple[Message, float]:
        """Receive a message and attached settings overlay.

        Receiving is a blocking operation. This function will contact
        the sender, wait for a message to be available, and receive and
        return it.

        If the port is not connected, then the default value will be
        returned if one was given, exactly as it was given. If no
        default was given then a RuntimeError will be raised.

        Args:
            port_name: The endpoint on which a message is to be
                    received.
            slot: The slot to receive the message on, if any.
            default: A message to return if this port is not connected.

        Returns:
            The received message, with message.settings holding
            the settings overlay. The settings attribute is
            guaranteed to not be None. Secondly, the saved_until
            metadata field from the received message.

        Raises:
            RuntimeError: If no default was given and the port is not
                connected.
        """
        if slot is None:
            port_and_slot = port_name
            slot_list = []      # type: List[int]
        else:
            port_and_slot = f"{port_name}[{slot}]"
            slot_list = [slot]
        _logger.debug('Waiting for message on {}'.format(port_and_slot))

        recv_endpoint = self.__get_endpoint(port_name, slot_list)

        if not self._peer_manager.is_connected(recv_endpoint.port):
            if default is None:
                raise RuntimeError(('Tried to receive on port "{}", which is'
                                    ' disconnected, and no default value was'
                                    ' given. Either specify a default, or'
                                    ' connect a sending component to this'
                                    ' port.').format(port_name))
            _logger.debug(
                    'No message received on {} as it is not connected'.format(
                        port_name))
            return default, float('-inf')

        if port_name in self._ports:
            port = self._ports[port_name]
        else:
            # it's muscle_settings_in here, because we check for unknown
            # user ports in Instance already, and we don't have any other
            # built-in automatic ports.
            port = self._muscle_settings_in

        profile_event = self._profiler.start(ProfileEventType.RECEIVE, port,
                                             None, slot, None)

        # peer_manager already checks that there is at most one snd_endpoint
        # connected to the port we receive on
        snd_endpoint = self._peer_manager.get_peer_endpoints(
                recv_endpoint.port, slot_list)[0]
        client = self.__get_client(snd_endpoint.instance())
        mpp_message_bytes = client.receive(recv_endpoint.ref())
        mpp_message = MPPMessage.from_bytes(mpp_message_bytes)

        if mpp_message.port_length is not None:
            if port.is_resizable():
                port.set_length(mpp_message.port_length)

        if isinstance(mpp_message.data, ClosePort):
            port.set_closed(slot)

        message = Message(
                mpp_message.timestamp, mpp_message.next_timestamp,
                mpp_message.data, mpp_message.settings_overlay)

        profile_event.stop()
        if port.is_vector():
            profile_event.port_length = port.get_length()
        profile_event.message_size = len(mpp_message_bytes)

        expected_message_number = port.get_num_messages(slot)
        if expected_message_number != mpp_message.message_number:
            if (expected_message_number - 1 == mpp_message.message_number and
                    port.is_resuming(slot)):
                _logger.debug(f'Discarding received message on {port_and_slot}'
                              ': resuming from weakly consistent snapshot')
                port.set_resumed(slot)
                return self.receive_message(port_name, slot, default)
            raise RuntimeError(f'Received message on {port_and_slot} with'
                               ' unexpected message number'
                               f' {mpp_message.message_number}. Was expecting'
                               f' {expected_message_number}. Are you resuming'
                               ' from an inconsistent snapshot?')
        port.increment_num_messages(slot)

        _logger.debug('Received message on {}'.format(port_and_slot))
        if isinstance(mpp_message.data, ClosePort):
            _logger.debug('Port {} is now closed'.format(port_and_slot))

        return message, mpp_message.saved_until

    def close_port(self, port_name: str, slot: Optional[int] = None
                   ) -> None:
        """Closes the given port.

        This signals to any connected instance that no more messages
        will be sent on this port, which it can use to decide whether
        to shut down or continue running.

        Args:
            port_name: The name of the port to close.
        """
        message = Message(float('inf'), None, ClosePort(), Settings())
        if slot is None:
            _logger.debug('Closing port {}'.format(port_name))
        else:
            _logger.debug('Closing port {}[{}]'.format(port_name, slot))
        self.send_message(port_name, message, slot)

    def shutdown(self) -> None:
        """Shuts down the Communicator, closing connections.
        """
        for client in self._clients.values():
            client.close()

        self._post_office.wait_for_receivers()

        for server in self._servers:
            server.close()

    def restore_message_counts(self, port_message_counts: Dict[str, List[int]]
                               ) -> None:
        """Restore message counts on all ports
        """
        for port_name, num_messages in port_message_counts.items():
            if port_name == "muscle_settings_in":
                self._muscle_settings_in.restore_message_counts(num_messages)
            elif port_name in self._ports:
                self._ports[port_name].restore_message_counts(num_messages)
            else:
                raise RuntimeError(f'Unknown port {port_name} in snapshot.'
                                   ' Have your port definitions changed since'
                                   ' the snapshot was taken?')

    def get_message_counts(self) -> Dict[str, List[int]]:
        """Get message counts for all ports on the communicator
        """
        port_message_counts = {port_name: port.get_message_counts()
                               for port_name, port in self._ports.items()}
        port_message_counts["muscle_settings_in"] = \
            self._muscle_settings_in.get_message_counts()
        return port_message_counts

    def __instance_id(self) -> Reference:
        """Returns our complete instance id.
        """
        return self._kernel + self._index

    def __ports_from_declared(self) -> Dict[str, Port]:
        """Derives port definitions from supplied declaration.
        """
        ports = dict()
        declared_ports = cast(Dict[Operator, List[str]], self._declared_ports)
        for operator, port_list in declared_ports.items():
            for port_desc in port_list:
                port_name, is_vector = self.__split_port_desc(port_desc)
                if port_name.startswith('muscle_'):
                    raise RuntimeError(('Port names starting with "muscle_"'
                                        ' are reserved for MUSCLE, please'
                                        ' rename port "{}"'.format(port_name)))
                port_id = Identifier(port_name)
                is_connected = self._peer_manager.is_connected(port_id)
                if is_connected:
                    peer_ports = self._peer_manager.get_peer_ports(port_id)
                    peer_port = peer_ports[0]
                    peer_ce = peer_port[:-1]
                    port_peer_dims = self._peer_manager.get_peer_dims(peer_ce)
                    for peer_port in peer_ports[1:]:
                        peer_ce = peer_port[:-1]
                        if port_peer_dims != self._peer_manager.get_peer_dims(
                                peer_ce):
                            port_strs = ', '.join(map(str, peer_ports))
                            raise RuntimeError(('Multicast port "{}" is'
                                                ' connected to peers with'
                                                ' different dimensions. All'
                                                ' peer components that this'
                                                ' port is connected to must'
                                                ' have the same multiplicity.'
                                                ' Connected to ports: {}.'
                                                ).format(port_name, port_strs))
                else:
                    port_peer_dims = []
                ports[port_name] = Port(
                        port_name, operator, is_vector, is_connected,
                        len(self._index), port_peer_dims)
        return ports

    def __ports_from_conduits(self, conduits: List[Conduit]
                              ) -> Dict[str, Port]:
        """Derives port definitions from conduits.

        Args:
            conduits: The list of conduits.
        """
        ports = dict()
        for conduit in conduits:
            if conduit.sending_component() == self._kernel:
                port_id = conduit.sending_port()
                operator = Operator.O_F
                port_peer_dims = self._peer_manager.get_peer_dims(
                        conduit.receiving_component())
            elif conduit.receiving_component() == self._kernel:
                port_id = conduit.receiving_port()
                operator = Operator.F_INIT
                port_peer_dims = self._peer_manager.get_peer_dims(
                        conduit.sending_component())
            else:
                continue

            ndims = max(0, len(port_peer_dims) - len(self._index))
            is_vector = (ndims == 1)
            is_connected = self._peer_manager.is_connected(port_id)
            if not str(port_id).startswith('muscle_'):
                ports[str(port_id)] = Port(
                        str(port_id), operator, is_vector, is_connected,
                        len(self._index), port_peer_dims)
        return ports

    def __settings_in_port(self, conduits: List[Conduit]) -> Port:
        """Creates a Port representing muscle_settings_in.

        Args:
            conduits: The list of conduits.
        """
        for conduit in conduits:
            if conduit.receiving_component() == self._kernel:
                port_id = conduit.receiving_port()
                if str(port_id) == 'muscle_settings_in':
                    return Port(str(port_id), Operator.F_INIT, False,
                                self._peer_manager.is_connected(port_id),
                                len(self._index),
                                self._peer_manager.get_peer_dims(
                                    conduit.sending_component()))
        return Port('muscle_settings_in', Operator.F_INIT, False, False,
                    len(self._index), [])

    def __get_client(self, instance: Reference) -> MPPClient:
        """Get or create a client to connect to the given instance.

        Args:
            instance: A reference to the instance to connect to.

        Returns:
            An existing or new MCP client.
        """
        if instance not in self._clients:
            locations = self._peer_manager.get_peer_locations(instance)
            _logger.info(f'Connecting to peer {instance} at {locations}')
            self._clients[instance] = MPPClient(locations)

        return self._clients[instance]

    def __get_endpoint(self, port_name: str, slot: List[int]) -> Endpoint:
        """Determines the endpoint on our side.

        Args:
            port_name: Name of the port to send or receive on.
            slot: Slot to send or receive on.
        """
        try:
            port = Identifier(port_name)
        except ValueError as e:
            raise ValueError('"{}" is not a valid port name: {}'.format(
                port_name, e))

        return Endpoint(self._kernel, self._index, port, slot)

    def __split_port_desc(self, port_desc: str) -> Tuple[str, bool]:
        """Split a port description into its name and dimensionality.

        Expects a port description of the form port_name or
        port_name[], and returns the port name and whether it is a
        vector port.

        Args:
            port_desc: A port description string, as above.
        """
        is_vector = False
        if port_desc.endswith('[]'):
            is_vector = True
            port_desc = port_desc[:-2]

        if port_desc.endswith('[]'):
            raise ValueError(('Port description "{}" is invalid: ports can'
                              ' have at most one dimension.').format(
                                  port_desc))

        return port_desc, is_vector



import struct

import msgpack
import numpy as np

from ymmsl import Reference, Settings

from libmuscle.grid import Grid
from libmuscle.mpp_message import MPPMessage


def test_create() -> None:
    sender = Reference('sender.port')
    receiver = Reference('receiver.port')
    timestamp = 10.0
    next_timestamp = 11.0
    settings_overlay = (6789).to_bytes(2, 'little', signed=True)
    message_number = 0
    saved_until = 1.6
    data = (12345).to_bytes(2, 'little', signed=True)
    msg = MPPMessage(
            sender, receiver, None, timestamp, next_timestamp,
            settings_overlay, message_number, saved_until, data)
    assert msg.sender == sender
    assert msg.receiver == receiver
    assert msg.port_length is None
    assert msg.timestamp == 10.0
    assert msg.next_timestamp == 11.0
    assert msg.settings_overlay == settings_overlay
    assert msg.message_number == 0
    assert msg.saved_until == 1.6
    assert msg.data == data


def test_grid_encode() -> None:
    sender = Reference('sender.port')
    receiver = Reference('receiver.port')
    timestamp = 10.0
    next_timestamp = 11.0

    array = np.array(
            [[[1.0, 2.0, 3.0],
              [4.0, 5.0, 6.0]],
             [[7.0, 8.0, 9.0],
              [10.0, 11.0, 12.0]]], np.float32)

    grid = Grid(array, ['x', 'y', 'z'])
    msg = MPPMessage(
            sender, receiver, None, timestamp, next_timestamp, Settings(),
            0, 1.0, grid)

    wire_data = msg.encoded()
    mcp_decoded = msgpack.unpackb(wire_data, raw=False)
    grid_decoded = msgpack.unpackb(mcp_decoded['data'].data, raw=False)

    assert grid_decoded['type'] == 'float32'
    assert grid_decoded['shape'] == [2, 2, 3]
    assert grid_decoded['order'] == 'la'
    next_value = 1.0
    for value in struct.iter_unpack('<f', grid_decoded['data']):
        assert value[0] == next_value
        next_value = next_value + 1.0

    assert grid_decoded['indexes'] == ['x', 'y', 'z']


def test_grid_decode() -> None:
    settings_data = msgpack.packb({}, use_bin_type=True)

    grid_buf = bytes(
            [1, 0, 0, 0,
             2, 0, 0, 0,
             3, 0, 0, 0,
             0, 0, 4, 0,
             0, 0, 5, 0,
             0, 0, 6, 0])

    grid_dict = {
            'type': 'int32',
            'shape': [2, 3],
            'order': 'la',
            'data': grid_buf,
            'indexes': []}
    grid_data = msgpack.packb(grid_dict, use_bin_type=True)

    msg_dict = {
            'sender': 'elem1.port1',
            'receiver': 'elem2.port2',
            'port_length': 0,
            'timestamp': 0.0,
            'next_timestamp': None,
            'settings_overlay': msgpack.ExtType(1, settings_data),
            'message_number': 0,
            'saved_until': 9.9,
            'data': msgpack.ExtType(2, grid_data)}

    wire_data = msgpack.packb(msg_dict, use_bin_type=True)

    msg = MPPMessage.from_bytes(wire_data)

    assert isinstance(msg.data, Grid)
    assert msg.data.array.dtype == np.int32
    assert msg.data.array.shape == (2, 3)
    assert msg.data.array.flags.c_contiguous
    assert msg.data.array[0, 0] == 1
    assert msg.data.array[0, 1] == 2
    assert msg.data.array[1, 1] == 5 * 65536
    assert msg.data.indexes is None

    grid_dict['order'] = 'fa'
    grid_data = msgpack.packb(grid_dict, use_bin_type=True)
    msg_dict['data'] = msgpack.ExtType(2, grid_data)
    wire_data = msgpack.packb(msg_dict, use_bin_type=True)
    msg = MPPMessage.from_bytes(wire_data)

    assert isinstance(msg.data, Grid)
    assert msg.data.array.dtype == np.int32
    assert msg.data.array.shape == (2, 3)
    assert msg.data.array.flags.f_contiguous
    assert msg.data.array[0, 0] == 1
    assert msg.data.array[0, 1] == 3
    assert msg.data.array[0, 2] == 5 * 65536
    assert msg.data.indexes is None


def test_grid_roundtrip() -> None:
    sender = Reference('sender.port')
    receiver = Reference('receiver.port')
    timestamp = 10.0
    next_timestamp = 11.0

    for order in ('C', 'F'):
        array = np.array(
                [[[1.0, 2.0, 3.0],
                  [4.0, 5.0, 6.0]],
                 [[7.0, 8.0, 9.0],
                  [10.0, 11.0, 12.0]]], np.float64, order=order)

        assert array[0, 0, 0] == 1.0

        grid = Grid(array, ['x', 'y', 'z'])
        msg = MPPMessage(
                sender, receiver, None, timestamp, next_timestamp, Settings(),
                0, 1.0, grid)

        wire_data = msg.encoded()
        msg_out = MPPMessage.from_bytes(wire_data)

        assert isinstance(msg_out.data, Grid)
        grid_out = msg_out.data
        assert grid_out.indexes == ['x', 'y', 'z']
        assert isinstance(grid_out.array, np.ndarray)
        assert grid_out.array.dtype == np.float64
        assert grid_out.array.shape == (2, 2, 3)
        assert grid_out.array.size == 12
        assert grid_out.array[1, 0, 1] == 8.0
        assert grid_out.array[0, 0, 2] == 3.0


def test_non_contiguous_grid_roundtrip() -> None:
    sender = Reference('sender.port')
    receiver = Reference('receiver.port')
    timestamp = 10.0
    next_timestamp = 11.0

    array = np.array(
            [[[1.0 + 0.125j, 2.0 + 0.25j, 3.0 + 0.375j],
              [4.0 + 0.4375j, 5.0 + 0.5j, 6.0 + 0.625j]],
             [[7.0 + 0.75j, 8.0 + 0.875j, 9.0 + 0.9375j],
              [10.0 + 0.10j, 11.0 + 0.11j, 12.0 + 0.12j]]], np.complex64)

    assert array.real[0, 0, 0] == 1.0
    assert array.imag[0, 0, 0] == 0.125

    grid = Grid(array.real, ['a', 'b', 'c'])
    msg = MPPMessage(
            sender, receiver, None, timestamp, next_timestamp, Settings(),
            0, 7.7, grid)

    wire_data = msg.encoded()
    msg_out = MPPMessage.from_bytes(wire_data)

    assert isinstance(msg_out.data, Grid)
    grid_out = msg_out.data
    assert isinstance(grid_out.array, np.ndarray)
    assert grid_out.array.dtype == np.float32
    assert grid_out.array.shape == (2, 2, 3)
    assert grid_out.array.size == 12
    assert grid_out.array[1, 0, 1] == 8.0
    assert grid_out.array[0, 0, 2] == 3.0



from libmuscle.mcp.tcp_transport_client import TcpTransportClient
from libmuscle.mcp.tcp_transport_server import TcpTransportServer


# These must be in order of preference, i.e. most efficient first
transport_client_types = [TcpTransportClient]


transport_server_types = [TcpTransportServer]



from enum import Enum


class RequestType(Enum):
    """Identifier for different types of request

    The MUSCLE Communication Protocol (MCP) defines a simple Remote Procedure
    Call protocol in which a request is sent to the server and a response is
    sent back to the calling client. In MCP, both of these are chunks of bytes.

    The MUSCLE Manager Protocol and MUSCLE Peer Protocol define the encoded
    messages sent in those chunks, using MsgPack encoding. To distinguish
    different kinds of requests, a request type identifier is used, as
    represented by this class.
    """
    # MUSCLE Manager Protocol
    REGISTER_INSTANCE = 1
    GET_PEERS = 2
    DEREGISTER_INSTANCE = 3
    GET_SETTINGS = 4
    SUBMIT_LOG_MESSAGE = 5
    SUBMIT_PROFILE_EVENTS = 6
    SUBMIT_SNAPSHOT = 7
    GET_CHECKPOINT_INFO = 8

    # MUSCLE Peer Protocol
    GET_NEXT_MESSAGE = 21


class ResponseType(Enum):
    """Identifier for different types of response

    MCP Responses may be of different kinds, as described by this class.
    These are currently only used by the MUSCLE Manager Protocol.
    """
    # MUSCLE Manager Protocol
    SUCCESS = 0
    ERROR = 1
    PENDING = 2



import numpy as np
import pytest

from libmuscle.grid import Grid


def test_grid() -> None:
    a = np.array([[1, 2, 3], [4, 5, 6]])
    _ = Grid(a)
    _ = Grid(a, ['x', 'y'])

    with pytest.raises(ValueError):
        _ = Grid(a, ['x'])

    with pytest.raises(ValueError):
        _ = Grid(a, ['x', 'y', 'z'])



from socket import SocketType


class SocketClosed(Exception):
    """Raised when trying to read from a socket that was closed.
    """
    pass


def recv_all(socket: SocketType, length: int) -> bytes:
    """Receive length bytes from a socket.

    Args:
        socket: Socket to receive on.
        length: Number of bytes to receive.

    Raises:
        SocketClosed: If the socket was closed by the peer.
        RuntimeError: If a read error occurred.
    """
    databuf = bytearray(length)
    received_count = 0
    while received_count < length:
        bytes_left = length - received_count
        received_now = socket.recv_into(
            memoryview(databuf)[received_count:], bytes_left)

        if received_now == 0:
            raise SocketClosed("Socket closed while receiving")

        if received_now == -1:
            raise RuntimeError("Error receiving")

        received_count += received_now

    return databuf


def send_int64(socket: SocketType, data: int) -> None:
    """Sends an int as a 64-bit signed little endian number.

    Args:
        socket: The socket to send on.
        data: The number to send.

    Raises:
        RuntimeError: If there was an error sending the data.
    """
    buf = data.to_bytes(8, byteorder='little')
    socket.sendall(buf)


def recv_int64(socket: SocketType) -> int:
    """Receives an int as a 64-bit signed little endian number.

    Args:
        socket: The socket to receive on.

    Raises:
        SocketClosed: If the socket was closed by the peer.
        RuntimeError: If a read error occurred.
    """
    buf = recv_all(socket, 8)
    return int.from_bytes(buf, 'little')



import logging
from typing import List
from libmuscle.communicator import Communicator, Endpoint, Message
from libmuscle.mpp_message import ClosePort, MPPMessage
from libmuscle.port import Port

from ymmsl import Conduit, Identifier, Operator, Reference, Settings

import pytest
from unittest.mock import patch, MagicMock


def test_endpoint() -> None:
    kernel = Reference('test.kernel')
    index = [42]
    port = Identifier('out')
    slot = [2]

    endpoint = Endpoint(kernel, index, port, slot)
    assert endpoint.kernel == kernel
    assert endpoint.index == index
    assert endpoint.port == port
    assert endpoint.slot == slot

    assert str(endpoint) == 'test.kernel[42].out[2]'


def test_endpoint_instance() -> None:
    endpoint = Endpoint(Reference('test.kernel'), [42], Identifier('port'),
                        [2])
    assert endpoint.instance() == 'test.kernel[42]'

    endpoint2 = Endpoint(Reference('test.kernel'), [], Identifier('port'), [])
    assert endpoint2.instance() == 'test.kernel'

    endpoint3 = Endpoint(Reference('test.kernel'), [], Identifier('port'), [3])
    assert endpoint3.instance() == 'test.kernel'


@pytest.fixture
def communicator() -> Communicator:
    instance_id = Reference('kernel')
    communicator = Communicator(instance_id, [13], None, MagicMock())
    communicator._peer_manager = MagicMock()
    pm = communicator._peer_manager
    pm.is_connected.return_value = True

    def gpp(x) -> List[Reference]:
        if 'out' in str(x):
            return [Reference('in')]
        return [Reference('out')]

    pm.get_peer_ports = gpp

    pm.get_peer_dims.return_value = []
    pm.get_peer_locations.return_value = ['direct:test']

    def gpe(p, s) -> List[Reference]:
        endpoint = MagicMock()
        endpoint.instance.return_value = Reference('other')
        if 'out' in str(p):
            endpoint.ref.return_value = Reference('other.in[13]')
        else:
            endpoint.ref.return_value = Reference('other.out')
        return [endpoint]

    pm.get_peer_endpoints = gpe

    communicator._ports = {
            'out': Port('out', Operator.O_I, False, True, 1, []),
            'in': Port('in', Operator.S, False, True, 1, [])}
    communicator._muscle_settings_in = \
        communicator._Communicator__settings_in_port([])
    yield communicator
    communicator.shutdown()


@pytest.fixture
def communicator2() -> Communicator:
    instance_id = Reference('other')
    communicator = Communicator(instance_id, [], None, MagicMock())
    communicator._peer_manager = MagicMock()
    pm = communicator._peer_manager
    pm.is_connected.return_value = True

    def gpp(x) -> List[Reference]:
        if 'out' in str(x):
            return [Reference('in')]
        return [Reference('out')]

    pm.get_peer_ports = gpp

    pm.get_peer_dims.return_value = []
    pm.get_peer_locations.return_value = ['direct:test']

    def gpe(p, s) -> List[Reference]:
        endpoint = MagicMock()
        endpoint.instance.return_value = Reference('kernel[13]')
        if 'out' in str(p):
            endpoint.ref.return_value = Reference('kernel[13].in')
        else:
            endpoint.ref.return_value = Reference('kernel[13].out')
        return [endpoint]

    pm.get_peer_endpoints = gpe

    communicator._ports = {
            'out': Port('out', Operator.O_I, True, True, 0, [20]),
            'in': Port('in', Operator.S, True, True, 0, [20])}
    communicator._muscle_settings_in = \
        communicator._Communicator__settings_in_port([])
    yield communicator
    communicator.shutdown()


@pytest.fixture
def communicator3() -> Communicator:
    instance_id = Reference('kernel')
    communicator = Communicator(instance_id, [], None, MagicMock())
    communicator._peer_manager = MagicMock()
    pm = communicator._peer_manager
    pm.is_connected.return_value = True

    def gpp(x) -> List[Reference]:
        if 'out' in str(x):
            return [Reference('in')]
        return [Reference('out')]

    pm.get_peer_ports = gpp

    pm.get_peer_dims.return_value = []
    pm.get_peer_locations.return_value = ['direct:test']

    def gpe(p, s) -> List[Reference]:
        endpoint = MagicMock()
        endpoint.instance.return_value = Reference('other')
        if 'out' in str(p):
            endpoint.ref.return_value = Reference('other.in[13]')
        else:
            endpoint.ref.return_value = Reference('other.out[13]')
        return [endpoint]

    pm.get_peer_endpoints = gpe

    communicator._ports = {
            'out': Port('out', Operator.O_I, True, True, 0, []),
            'in': Port('in', Operator.S, True, True, 0, [])}
    yield communicator
    communicator.shutdown()


def test_create_communicator(communicator) -> None:
    assert str(communicator._kernel) == 'kernel'
    assert communicator._index == [13]
    assert len(communicator._servers) == 1
    assert communicator._clients == {}
    assert communicator._post_office._outboxes == {}


def test_get_locations(communicator) -> None:
    assert len(communicator.get_locations()) == 1
    assert communicator.get_locations()[0].startswith('tcp:')


def test_connect() -> None:
    ref = Reference

    instance_id = Reference('kernel')
    conduits = [Conduit('kernel.out', 'other.in'),
                Conduit('other.out', 'kernel.in')]
    peer_dims = {ref('other'): [1]}
    peer_locations = {ref('other'): ['direct:test']}

    with patch('libmuscle.communicator.PeerManager') as pm_init:
        communicator = Communicator(instance_id, [13], None, MagicMock())

        communicator.connect(conduits, peer_dims, peer_locations)

        pm_init.assert_called_with(instance_id, [13], conduits, peer_dims,
                                   peer_locations)

        # check inferred ports
        ports = communicator._ports
        communicator.shutdown()
        assert ports['in'].name == Identifier('in')
        assert ports['in'].operator == Operator.F_INIT
        assert ports['in']._length is None

        assert ports['out'].name == Identifier('out')
        assert ports['out'].operator == Operator.O_F
        assert ports['out']._length is None


def test_connect_vector_ports(communicator) -> None:
    ref = Reference

    communicator._declared_ports = {
            Operator.F_INIT: ['in[]'],
            Operator.O_F: ['out1', 'out2[]']}

    conduits = [Conduit('other1.out', 'kernel.in'),
                Conduit('kernel.out1', 'other.in'),
                Conduit('kernel.out2', 'other3.in')]
    peer_dims = {
            ref('other1'): [20, 7],
            ref('other'): [25],
            ref('other3'): [20]}
    peer_locations = {
            ref('other'): ['direct:test'],
            ref('other1'): ['direct:test1'],
            ref('other3'): ['direct:test3']}

    communicator.connect(conduits, peer_dims, peer_locations)

    ports = communicator._ports
    assert ports['in'].name == Identifier('in')
    assert ports['in'].operator == Operator.F_INIT
    assert ports['in']._length == 7
    assert ports['in']._is_resizable is False

    assert ports['out1'].name == Identifier('out1')
    assert ports['out1'].operator == Operator.O_F
    assert ports['out1']._length is None

    assert ports['out2'].name == Identifier('out2')
    assert ports['out2'].operator == Operator.O_F
    assert ports['out2']._length == 0
    assert ports['out2']._is_resizable is True


def test_connect_multidimensional_ports(communicator) -> None:
    ref = Reference

    communicator._declared_ports = {
            Operator.F_INIT: ['in[][]']}

    conduits = [Conduit(ref('other.out'), ref('kernel.in'))]
    peer_dims = {ref('other'): [20, 7, 30]}
    peer_locations = {ref('other'): ['direct:test']}
    with pytest.raises(ValueError):
        communicator.connect(conduits, peer_dims, peer_locations)


def test_connect_inferred_ports(communicator) -> None:
    ref = Reference

    communicator._declared_ports = None

    conduits = [Conduit('other1.out', 'kernel.in'),
                Conduit('kernel.out1', 'other.in'),
                Conduit('kernel.out3', 'other2.in')]
    peer_dims = {
            ref('other1'): [20, 7],
            ref('other'): [25],
            ref('other2'): []}
    peer_locations = {
            ref('other'): ['direct:test'],
            ref('other1'): ['direct:test1'],
            ref('other2'): ['direct:test2']}

    communicator.connect(conduits, peer_dims, peer_locations)

    ports = communicator._ports
    assert ports['in'].name == Identifier('in')
    assert ports['in'].operator == Operator.F_INIT
    assert ports['in']._length == 7
    assert ports['in']._is_resizable is False

    assert ports['out1'].name == Identifier('out1')
    assert ports['out1'].operator == Operator.O_F
    assert ports['out1']._length is None

    assert ports['out3'].name == Identifier('out3')
    assert ports['out3'].operator == Operator.O_F
    assert ports['out3']._length is None


def test_send_message(communicator, message) -> None:
    communicator.send_message('out', message)

    assert 'other.in[13]' in communicator._post_office._outboxes
    msg_bytes = communicator._post_office._outboxes[
            'other.in[13]']._Outbox__queue.get()
    msg = MPPMessage.from_bytes(msg_bytes)
    assert msg.sender == 'kernel[13].out'
    assert msg.receiver == 'other.in[13]'
    assert msg.timestamp == 0.0
    assert msg.next_timestamp is None
    assert msg.settings_overlay == Settings()
    assert msg.message_number == 0
    assert msg.data == b'test'


def test_send_on_disconnected_port(communicator, message) -> None:
    communicator._peer_manager.is_connected.return_value = False
    communicator.send_message('not_connected', message)


def test_send_on_invalid_port(communicator, message) -> None:
    with pytest.raises(ValueError):
        communicator.send_message('[$Invalid_id', message)


def test_send_msgpack(communicator, message2) -> None:
    communicator.send_message('out', message2)

    assert 'other.in[13]' in communicator._post_office._outboxes
    msg_bytes = communicator._post_office._outboxes[
            'other.in[13]']._Outbox__queue.get()
    msg = MPPMessage.from_bytes(msg_bytes)
    assert msg.sender == 'kernel[13].out'
    assert msg.receiver == 'other.in[13]'
    assert msg.settings_overlay == Settings()
    assert msg.message_number == 0
    assert msg.data == {'test': 17}


def test_send_message_with_slot(communicator2, message) -> None:
    communicator2.send_message('out', message, 13)

    assert 'kernel[13].in' in \
        communicator2._post_office._outboxes
    msg_bytes = communicator2._post_office._outboxes[
            'kernel[13].in']._Outbox__queue.get()
    msg = MPPMessage.from_bytes(msg_bytes)
    assert msg.sender == 'other.out[13]'
    assert msg.receiver == 'kernel[13].in'
    assert msg.settings_overlay == Settings()
    assert msg.message_number == 0
    assert msg.data == b'test'


def test_send_message_resizable(communicator3, message) -> None:
    with pytest.raises(RuntimeError):
        communicator3.send_message('out', message, 13)

    communicator3.get_port('out').set_length(20)
    communicator3.send_message('out', message, 13)

    assert 'other.in[13]' in communicator3._post_office._outboxes
    msg_bytes = communicator3._post_office._outboxes[
            'other.in[13]']._Outbox__queue.get()
    msg = MPPMessage.from_bytes(msg_bytes)
    assert msg.sender == 'kernel.out[13]'
    assert msg.receiver == 'other.in[13]'
    assert msg.port_length == 20


def test_send_message_with_settings(communicator, message) -> None:
    message.settings['test2'] = 'testing'
    communicator.send_message('out', message)

    assert 'other.in[13]' in communicator._post_office._outboxes
    msg_bytes = communicator._post_office._outboxes[
            'other.in[13]']._Outbox__queue.get()
    msg = MPPMessage.from_bytes(msg_bytes)
    assert msg.sender == 'kernel[13].out'
    assert msg.receiver == 'other.in[13]'
    assert msg.settings_overlay.as_ordered_dict() == {'test2': 'testing'}
    assert msg.message_number == 0
    assert msg.data == b'test'


def test_send_settings(communicator, message) -> None:
    message.data = Settings()
    message.data['test1'] = 'testing'
    communicator.send_message('out', message)

    assert 'other.in[13]' in communicator._post_office._outboxes
    msg_bytes = communicator._post_office._outboxes[
            'other.in[13]']._Outbox__queue.get()
    msg = MPPMessage.from_bytes(msg_bytes)
    assert msg.sender == 'kernel[13].out'
    assert msg.receiver == 'other.in[13]'
    assert msg.settings_overlay == Settings()
    assert msg.message_number == 0
    assert msg.data == Settings({'test1': 'testing'})


def test_close_port(communicator) -> None:
    communicator.close_port('out')

    assert 'other.in[13]' in communicator._post_office._outboxes
    msg_bytes = communicator._post_office._outboxes[
            'other.in[13]']._Outbox__queue.get()
    msg = MPPMessage.from_bytes(msg_bytes)
    assert msg.sender == 'kernel[13].out'
    assert msg.receiver == 'other.in[13]'
    assert msg.timestamp == float('inf')
    assert msg.next_timestamp is None
    assert msg.settings_overlay == Settings()
    assert msg.message_number == 0
    assert isinstance(msg.data, ClosePort)


def test_receive_message(communicator) -> None:
    client_mock = MagicMock()
    client_mock.receive.return_value = MPPMessage(
            Reference('other.out[13]'), Reference('kernel[13].in'),
            None, 0.0, None, Settings({'test1': 12}), 0, 2.0,
            b'test').encoded()
    get_client_mock = MagicMock(return_value=client_mock)
    communicator._Communicator__get_client = get_client_mock
    communicator._profiler = MagicMock()

    msg, last_saved = communicator.receive_message('in')

    get_client_mock.assert_called_with(Reference('other'))
    client_mock.receive.assert_called_with(Reference('kernel[13].in'))
    assert msg.data == b'test'
    assert msg.settings['test1'] == 12
    assert last_saved == 2.0


def test_receive_message_default(communicator) -> None:
    communicator._peer_manager.is_connected.return_value = False
    default_msg = Message(3.0, 4.0, 'test', Settings())
    msg, last_saved = communicator.receive_message(
            'not_connected', default=default_msg)
    assert msg.timestamp == 3.0
    assert msg.next_timestamp == 4.0
    assert msg.data == 'test'
    assert len(msg.settings) == 0
    assert last_saved == float('-inf')


def test_receive_message_no_default(communicator) -> None:
    communicator._peer_manager.is_connected.return_value = False
    with pytest.raises(RuntimeError):
        communicator.receive_message('not_connected')


def test_receive_on_invalid_port(communicator) -> None:
    with pytest.raises(ValueError):
        communicator.receive_message('@$Invalid_id')


def test_receive_msgpack(communicator) -> None:
    client_mock = MagicMock()
    client_mock.receive.return_value = MPPMessage(
            Reference('other.out[13]'), Reference('kernel[13].in'),
            None, 0.0, None, Settings({'test1': 12}), 0, 1.0,
            {'test': 13}).encoded()
    get_client_mock = MagicMock(return_value=client_mock)
    communicator._Communicator__get_client = get_client_mock
    communicator._profiler = MagicMock()

    msg, last_saved = communicator.receive_message('in')

    get_client_mock.assert_called_with(Reference('other'))
    client_mock.receive.assert_called_with(Reference('kernel[13].in'))
    assert msg.data == {'test': 13}
    assert last_saved == 1.0


def test_receive_with_slot(communicator2) -> None:
    client_mock = MagicMock()
    client_mock.receive.return_value = MPPMessage(
            Reference('kernel[13].out'), Reference('other.in[13]'),
            None, 0.0, None, Settings({'test': 'testing'}), 0, 3.0,
            b'test').encoded()
    get_client_mock = MagicMock(return_value=client_mock)
    communicator2._Communicator__get_client = get_client_mock
    communicator2._profiler = MagicMock()

    msg, last_saved = communicator2.receive_message('in', 13)

    get_client_mock.assert_called_with(Reference('kernel[13]'))
    client_mock.receive.assert_called_with(Reference('other.in[13]'))
    assert msg.data == b'test'
    assert msg.settings['test'] == 'testing'
    assert last_saved == 3.0


def test_receive_message_resizable(communicator3) -> None:
    client_mock = MagicMock()
    client_mock.receive.return_value = MPPMessage(
            Reference('other.out[13]'), Reference('kernel.in[13]'),
            20, 0.0, None, Settings({'test': 'testing'}), 0, 12.3,
            b'test').encoded()
    get_client_mock = MagicMock(return_value=client_mock)
    communicator3._Communicator__get_client = get_client_mock
    communicator3._profiler = MagicMock()

    msg, last_saved = communicator3.receive_message('in', 13)

    get_client_mock.assert_called_with(Reference('other'))
    client_mock.receive.assert_called_with(Reference('kernel.in[13]'))
    assert msg.data == b'test'
    assert communicator3.get_port('in').get_length() == 20
    assert last_saved == 12.3


def test_receive_with_settings(communicator) -> None:
    client_mock = MagicMock()
    client_mock.receive.return_value = MPPMessage(
            Reference('other.out[13]'), Reference('kernel[13].in'),
            None, 0.0, None, Settings({'test2': 3.1}), 0, 0.1,
            b'test').encoded()
    get_client_mock = MagicMock(return_value=client_mock)
    communicator._Communicator__get_client = get_client_mock
    communicator._profiler = MagicMock()

    msg, last_saved = communicator.receive_message('in')

    get_client_mock.assert_called_with(Reference('other'))
    client_mock.receive.assert_called_with(Reference('kernel[13].in'))
    assert msg.data == b'test'
    assert msg.settings['test2'] == 3.1
    assert last_saved == 0.1


def test_receive_msgpack_with_slot_and_settings(communicator2) -> None:
    client_mock = MagicMock()
    client_mock.receive.return_value = MPPMessage(
            Reference('kernel[13].out'), Reference('other.in[13]'),
            None, 0.0, 1.0,
            Settings({'test': 'testing'}), 0, 1.0, 'test').encoded()
    get_client_mock = MagicMock(return_value=client_mock)
    communicator2._Communicator__get_client = get_client_mock
    communicator2._profiler = MagicMock()

    msg, last_saved = communicator2.receive_message('in', 13)

    get_client_mock.assert_called_with(Reference('kernel[13]'))
    client_mock.receive.assert_called_with(Reference('other.in[13]'))
    assert msg.data == 'test'
    assert msg.settings['test'] == 'testing'
    assert last_saved == 1.0


def test_receive_settings(communicator) -> None:
    client_mock = MagicMock()
    client_mock.receive.return_value = MPPMessage(
            Reference('other.out[13]'), Reference('kernel[13].in'),
            None, 0.0, None, Settings({'test1': 12}), 0, 1.0,
            Settings({'test': 13})).encoded()
    get_client_mock = MagicMock(return_value=client_mock)
    communicator._Communicator__get_client = get_client_mock
    communicator._profiler = MagicMock()

    msg, last_saved = communicator.receive_message('in')

    get_client_mock.assert_called_with(Reference('other'))
    client_mock.receive.assert_called_with(Reference('kernel[13].in'))
    assert isinstance(msg.data, Settings)
    assert msg.data['test'] == 13
    assert last_saved == 1.0


def test_receive_close_port(communicator) -> None:
    client_mock = MagicMock()
    client_mock.receive.return_value = MPPMessage(
            Reference('other.out[13]'), Reference('kernel[13].in'),
            None, 0.0, None, Settings(), 0, 0.1, ClosePort()).encoded()
    get_client_mock = MagicMock(return_value=client_mock)
    communicator._Communicator__get_client = get_client_mock
    communicator._profiler = MagicMock()

    msg, _ = communicator.receive_message('in')

    assert isinstance(msg.data, ClosePort)


def test_get_message(communicator, message) -> None:
    communicator.send_message('out', message, None, 2.0)
    ref_message = MPPMessage(
            Reference('kernel[13].out'), Reference('other.in[13]'),
            None, 0.0, None, Settings(), 0, 2.0, b'test').encoded()
    assert communicator._post_office.get_message(
            'other.in[13]') == ref_message


def test_port_message_counts(communicator, message) -> None:
    communicator.send_message('out', message)
    msg_counts = communicator.get_message_counts()
    assert msg_counts == {'out': [1],
                          'in': [0],
                          'muscle_settings_in': [0]}

    communicator.restore_message_counts({'out': [3],
                                         'in': [2],
                                         'muscle_settings_in': [4]})
    communicator.send_message('out', message)
    msg_counts = communicator.get_message_counts()
    assert msg_counts == {'out': [4],
                          'in': [2],
                          'muscle_settings_in': [4]}

    # empty post office
    communicator._post_office.get_message('other.in[13]')
    communicator._post_office.get_message('other.in[13]')

    with pytest.raises(RuntimeError):
        communicator.restore_message_counts({"x?invalid_port": 3})


def test_vector_port_message_counts(communicator2, message) -> None:
    msg_counts = communicator2.get_message_counts()
    assert msg_counts == {'out': [0] * 20,
                          'in': [0] * 20,
                          'muscle_settings_in': [0]}

    communicator2.send_message('out', message, 13)
    msg_counts = communicator2.get_message_counts()
    assert msg_counts == {'out': [0] * 13 + [1] + [0] * 6,
                          'in': [0] * 20,
                          'muscle_settings_in': [0]}

    communicator2.restore_message_counts({'out': list(range(20)),
                                          'in': list(range(20)),
                                          'muscle_settings_in': [4]})
    communicator2.send_message('out', message, 13)
    msg_counts = communicator2.get_message_counts()
    assert msg_counts == {'out': list(range(13)) + [14] + list(range(14, 20)),
                          'in': list(range(20)),
                          'muscle_settings_in': [4]}

    # empty post office
    communicator2._post_office.get_message('kernel[13].in')
    communicator2._post_office.get_message('kernel[13].in')


def test_port_count_validation(communicator):
    client_mock = MagicMock()
    client_mock.receive.return_value = MPPMessage(
            Reference('other.out[13]'), Reference('kernel[13].in'),
            None, 0.0, None, Settings({'test1': 12}), 0, 7.6,
            b'test').encoded()
    get_client_mock = MagicMock(return_value=client_mock)
    communicator._Communicator__get_client = get_client_mock
    communicator._profiler = MagicMock()

    communicator.receive_message('in')
    assert communicator.get_message_counts()['in'] == [1]

    with pytest.raises(RuntimeError):
        # the message received has message_number = 0 again
        communicator.receive_message('in')


def test_port_discard_error_on_resume(caplog, communicator):
    client_mock = MagicMock()
    client_mock.receive.return_value = MPPMessage(
            Reference('other.out[13]'), Reference('kernel[13].in'),
            None, 0.0, None, Settings({'test1': 12}), 1, 2.3,
            b'test').encoded()
    get_client_mock = MagicMock(return_value=client_mock)
    communicator._Communicator__get_client = get_client_mock
    communicator._profiler = MagicMock()

    communicator.restore_message_counts({'out': [0],
                                         'in': [2],
                                         'muscle_settings_in': [0]})
    for port in communicator._ports.values():
        assert port._is_resuming == [True]
        assert port.is_resuming(None)

    # In the next block, the first message with message_number=1 is discarded.
    # The RuntimeError is raised when 'receiving' the second message with
    # message_number=1
    with caplog.at_level(logging.DEBUG, 'libmuscle.communicator'):
        with pytest.raises(RuntimeError):
            communicator.receive_message('in')
        # records 0, 2 and 3 are debug logs for starting/receiving on port
        assert 'Discarding received message' in caplog.records[1].message


def test_port_discard_success_on_resume(caplog, communicator):
    client_mock = MagicMock()
    client_mock.receive.side_effect = [MPPMessage(
            Reference('other.out[13]'), Reference('kernel[13].in'),
            None, 0.0, None, Settings({'test1': 12}), message_number, 1.0,
            {'this is message': message_number}).encoded()
            for message_number in [1, 2]]
    get_client_mock = MagicMock(return_value=client_mock)
    communicator._Communicator__get_client = get_client_mock
    communicator._profiler = MagicMock()

    communicator.restore_message_counts({'out': [0],
                                         'in': [2],
                                         'muscle_settings_in': [0]})
    for port in communicator._ports.values():
        assert port._is_resuming == [True]
        assert port.is_resuming(None)

    with caplog.at_level(logging.DEBUG, 'libmuscle.communicator'):
        msg, _ = communicator.receive_message('in')
        # records 0, 2 and 3 are debug logs for starting/receiving on port
        assert 'Discarding received message' in caplog.records[1].message
    # message_number=1 should be discarded:
    assert msg.data == {'this is message': 2}
    assert communicator.get_message_counts()['in'] == [3]



from unittest.mock import MagicMock

from libmuscle.mcp.tcp_transport_client import TcpTransportClient
from libmuscle.mcp.tcp_transport_server import TcpTransportServer


def test_tcp_transport():
    request = b'request'
    response = b'response'

    def handle_request(request: bytes) -> bytes:
        assert request == b'request'
        return response

    handler = MagicMock()
    handler.handle_request = handle_request

    # create server
    server = TcpTransportServer(handler)

    # create client
    server_location = server.get_location()
    assert TcpTransportClient.can_connect_to(server_location)
    client = TcpTransportClient(server_location)

    response2 = client.call(request)
    assert response == response2

    client.close()
    server.close()



import logging
from threading import Thread
from typing import Union
from multiprocessing import Queue
import queue

from ymmsl import Configuration

from libmuscle.manager.instantiator import (
        CancelAllRequest, CrashedResult, InstantiatorRequest,
        InstantiationRequest, ProcessStatus, ShutdownRequest)
from libmuscle.manager.qcgpj_instantiator import Process, QCGPJInstantiator
from libmuscle.manager.run_dir import RunDir
from libmuscle.planner.planner import Planner, Resources


_logger = logging.getLogger(__name__)


class LogHandlingThread(Thread):
    """Pumps log records from a queue.

    This gets log records from a queue and sends them to the logging
    system.
    """
    def __init__(self, queue: Queue) -> None:
        """Creates a LogHandlingThread.

        Args:
            queue: Queue to get log records from.
        """
        super().__init__()
        self._queue = queue
        self._shutting_down = False

    def shutdown(self) -> None:
        """Process remaining records and stop the thread."""
        self._shutting_down = True

    def run(self) -> None:
        """The thread's entry point."""
        while True:
            try:
                record = self._queue.get(True, 0.1)
                logger = logging.getLogger(record.name)
                logger.handle(record)
            except queue.Empty:
                if self._shutting_down:
                    break


_ResultType = Union[Process, CrashedResult]


class InstanceManager:
    """Instantiates and manages running instances"""
    def __init__(
            self, configuration: Configuration, run_dir: RunDir) -> None:
        """Create a ProcessManager.

        Args:
            configuration: The global configuration
            run_dir: Directory to run in
        """
        self._configuration = configuration
        self._run_dir = run_dir

        self._resources_in = Queue()    # type: Queue[Resources]
        self._requests_out = Queue()    # type: Queue[InstantiatorRequest]
        self._results_in = Queue()      # type: Queue[_ResultType]
        self._log_records_in = Queue()  # type: Queue[logging.LogRecord]

        self._instantiator = QCGPJInstantiator(
                self._resources_in, self._requests_out, self._results_in,
                self._log_records_in, self._run_dir.path)
        self._instantiator.start()

        self._log_handler = LogHandlingThread(self._log_records_in)
        self._log_handler.start()

        self._planner = Planner(self._resources_in.get())
        self._num_running = 0

    def set_manager_location(self, location: str) -> None:
        """Sets the network location of the manager.

        Call this before starting any instances so that the location
        can be passed to them.

        Args:
            location: The network location (e.g. localhost:5000)
        """
        self._manager_location = location

    def start_all(self) -> None:
        """Starts all the instances of the model."""
        allocations = self._planner.allocate_all(self._configuration)
        for instance, resources in allocations.items():
            _logger.info(f'Planned {instance} on {resources}')

        components = {c.name: c for c in self._configuration.model.components}
        for instance, resources in allocations.items():
            component = components[instance.without_trailing_ints()]
            if component.implementation is None:
                _logger.warning(
                        f'No implementation specified for {component.name}'
                        ', not starting it.')
                continue
            implementation = self._configuration.implementations[
                    component.implementation]
            implementation.env['MUSCLE_MANAGER'] = self._manager_location
            idir = self._run_dir.add_instance_dir(instance)
            workdir = idir / 'workdir'
            workdir.mkdir()
            stdout_path = idir / 'stdout.txt'
            stderr_path = idir / 'stderr.txt'

            request = InstantiationRequest(
                    instance, implementation,
                    self._configuration.resources[component.name],
                    resources, idir, workdir, stdout_path, stderr_path)
            _logger.info(f'Instantiating {instance} on {resources}')
            self._requests_out.put(request)
            self._num_running += 1

    def wait(self) -> bool:
        """Waits for all instances to be done."""
        all_seemingly_okay = True
        while self._num_running > 0:
            result = self._results_in.get()

            if isinstance(result, CrashedResult):
                _logger.error(
                    'Instantiator crashed. This should not happen, please file'
                    ' a bug report.')
                return False

            if result.exit_code != 0:
                if result.status == ProcessStatus.CANCELED:
                    _logger.info(
                            f'Instance {result.instance} was shut down by'
                            f' MUSCLE3 because an error occurred elsewhere')
                else:
                    _logger.error(
                            f'Instance {result.instance} quit with error'
                            f' {result.exit_code}')
                    _logger.error(
                            'Output may be found in'
                            f' {self._run_dir.instance_dir(result.instance)}')
                    if all_seemingly_okay:
                        self._requests_out.put(CancelAllRequest())
                        all_seemingly_okay = False
            else:
                if result.status == ProcessStatus.CANCELED:
                    _logger.info(
                            f'Instance {result.instance} was not started'
                            f' because of an error elsewhere')
                else:
                    _logger.debug(f'Instance {result.instance} finished')
                    _logger.debug(f'States: {result.status}')
                    _logger.debug(f'Exit code: {result.exit_code}')
                    _logger.debug(f'Error msg: {result.error_msg}')

            self._num_running -= 1
        return all_seemingly_okay

    def shutdown(self) -> None:
        """Shuts down the backend.

        This will wait for any processes still running before shutting
        down and returning.
        """
        _logger.debug('Shutting down instance manager')
        self._requests_out.put(CancelAllRequest())
        self._requests_out.put(ShutdownRequest())
        self._instantiator.join()
        self._log_handler.shutdown()
        self._log_handler.join()
        _logger.debug('Instance manager shut down cleanly')



import multiprocessing as mp

import ymmsl

from libmuscle.logging import LogLevel, LogMessage, Timestamp
from libmuscle.manager.manager import Manager
from libmuscle.mmp_client import MMPClient


def do_logging_test(caplog):
    ymmsl_text = (
            'ymmsl_version: v0.1\n'
            'model:\n'
            '  name: test_model\n'
            '  components:\n'
            '    macro: macro_implementation\n'
            '    micro:\n'
            '      implementation: micro_implementation\n'
            '      multiplicity: [10]\n'
            '  conduits:\n'
            '    macro.out: micro.in\n'
            '    micro.out: macro.in\n'
            'settings:\n'
            '  test1: 13\n'
            '  test2: 13.3\n'
            '  test3: testing\n'
            '  test4: True\n'
            '  test5: [2.3, 5.6]\n'
            '  test6:\n'
            '    - [1.0, 2.0]\n'
            '    - [3.0, 1.0]\n'
            )

    # create server
    ymmsl_doc = ymmsl.load(ymmsl_text)
    manager = Manager(ymmsl_doc)

    # create client
    client = MMPClient(manager.get_server_location())
    message = LogMessage(
            instance_id='test_logging',
            timestamp=Timestamp(2.0),
            level=LogLevel.DEBUG,
            text='Integration testing')

    # log and check
    client.submit_log_message(message)
    for rec in caplog.records:
        if rec.name == 'instances.test_logging':
            assert rec.time_stamp == '1970-01-01T00:00:02Z'
            assert rec.levelname == 'DEBUG'
            assert rec.message == 'Integration testing'
            break

    client.close()
    manager.stop()


def test_logging(log_file_in_tmpdir, caplog):
    process = mp.Process(target=do_logging_test, args=(caplog,))
    process.start()
    process.join()
    assert process.exitcode == 0



from enum import auto, Enum


class APIPhase(Enum):
    """Different phases that the user code traverses.

    These values describe different regions that the model code can be
    in for the case where checkpointing is implemented. By tracking
    the phase that the model should be in, we can detect incorrect API
    usage.

    This does not match the yMMSL operators, as it is more
    fine-grained and concerns checkpointing, which is not represented
    in the SEL.

    Note that AFTER_REUSE_INSTANCE and BEFORE_RESUMING refer to the
    same place in the code. AFTER_REUSE_INSTANCE is used when we
    don't know yet if the code has checkpointing support, and so we
    don't know whether the next call is to resuming() or to
    reuse_instance(). Once a checkpointing function has been called,
    we know that we should expect resume() after reuse_instance() and
    we use BEFORE_RESUMING accordingly.
    """
    BEFORE_FIRST_REUSE_INSTANCE = auto()
    """Before the first time calling reuse_instance"""

    BEFORE_REUSE_INSTANCE = auto()
    """Before calling reuse_instance"""

    BEFORE_RESUMING = auto()
    """Between reuse_instance and resuming"""

    BEFORE_LOAD_SNAPSHOT = auto()
    """Between resuming and load_snapshot"""

    BEFORE_SHOULD_INIT = auto()
    """After resuming, before should_init"""

    BEFORE_SHOULD_SAVE_SNAPSHOT = auto()
    """Between should_init and should_save*"""

    BEFORE_SAVE_SNAPSHOT = auto()
    """Between should_save_snapshot and save_snapshot"""

    BEFORE_SAVE_FINAL_SNAPSHOT = auto()
    """Between should_save_final_snapshot and save_final_snapshot"""

    AFTER_REUSE_LOOP = auto()
    """After the final call to reuse_instance()"""


class APIGuard:
    """Keeps track of and checks in which phase the model is.

    The verify_* functions are called when the corresponding function
    on Instance is called, to check that we're in the right phase. They
    raise a RuntimeError if there's a problem. The *_done functions are
    called to signal that the corresponding function finished
    successfully, and that we are moving on to the next phase.
    """
    def __init__(self, uses_checkpointing: bool) -> None:
        """Create an APIPhaseTracker.

        This starts the tracker in BEFORE_FIRST_REUSE_INSTANCE.
        """
        self._phase = APIPhase.BEFORE_FIRST_REUSE_INSTANCE
        self._uses_checkpointing = uses_checkpointing

    def _generic_error_messages(self, verify_phase: str) -> None:
        if self._phase in (
                APIPhase.BEFORE_FIRST_REUSE_INSTANCE,
                APIPhase.AFTER_REUSE_LOOP):
            msg = f'Please only call {verify_phase} inside the reuse loop.'
        elif self._phase == APIPhase.BEFORE_REUSE_INSTANCE:
            msg = (
                    'Please do not call {verify_phase} after'
                    ' should_save_final_snapshot. should_save_final_snapshot'
                    ' should be at the end of the reuse loop.')
        elif self._phase == APIPhase.BEFORE_RESUMING:
            msg = 'Inside the reuse loop you must call resuming first.'
        elif self._phase == APIPhase.BEFORE_LOAD_SNAPSHOT:
            msg = (
                    'If resuming returns True, then you must call'
                    ' load_snapshot first.')
        elif self._phase == APIPhase.BEFORE_SHOULD_INIT:
            msg = 'After calling resuming, you must call should_init first.'
        elif self._phase == APIPhase.BEFORE_SHOULD_SAVE_SNAPSHOT:
            msg = 'You must call save_snapshot or save_final_snapshot first.'
        elif self._phase == APIPhase.BEFORE_SAVE_SNAPSHOT:
            msg = (
                    'If should_save_snapshot returns True, then you must'
                    ' call save_snapshot first.')
        elif self._phase == APIPhase.BEFORE_SAVE_FINAL_SNAPSHOT:
            msg = (
                    'If should_save_final_snapshot returns True, then you'
                    ' must call save_final_snapshot first.')
        else:
            return
        raise RuntimeError(msg)

    def verify_reuse_instance(self) -> None:
        """Check reuse_instance()"""
        if self._phase not in (
                APIPhase.BEFORE_REUSE_INSTANCE,
                APIPhase.BEFORE_FIRST_REUSE_INSTANCE):
            raise RuntimeError(
                    'We reached the end of the reuse loop without checking'
                    ' if a snapshot should be saved. Please add at least'
                    ' a should_save_final_snapshot and save_final_snapshot.')

    def reuse_instance_done(self, reusing: bool) -> None:
        """Update phase on successful reuse_instance().

        Args:
            reusing: Whether we are reusing or not.
        """
        if not reusing:
            self._phase = APIPhase.AFTER_REUSE_LOOP
        else:
            if self._uses_checkpointing:
                self._phase = APIPhase.BEFORE_RESUMING
            else:
                self._phase = APIPhase.BEFORE_REUSE_INSTANCE

    def verify_resuming(self) -> None:
        """Check resuming()"""
        if not self._uses_checkpointing:
            raise RuntimeError(
                    'Please add the flag'
                    ' :attr:`InstanceFlag.USES_CHECKPOINT_API` to your'
                    ' instance to use the MUSCLE3 checkpointing API.')
        if self._phase != APIPhase.BEFORE_RESUMING:
            raise RuntimeError(
                    'Please call resuming() only as the first thing in the'
                    ' reuse loop.')

    def resuming_done(self, resuming: bool) -> None:
        """Update phase on successful resuming().

        Args:
            resuming: Whether we're resuming or not.
        """
        self._uses_checkpointing = True
        if resuming:
            self._phase = APIPhase.BEFORE_LOAD_SNAPSHOT
        else:
            self._phase = APIPhase.BEFORE_SHOULD_INIT

    def verify_load_snapshot(self) -> None:
        """Check load_snapshot()"""
        if self._phase != APIPhase.BEFORE_LOAD_SNAPSHOT:
            raise RuntimeError(
                    'Please check that we are resuming by calling resuming()'
                    ' before calling load_snapshot()')

    def load_snapshot_done(self) -> None:
        """Update phase on successful load_snapshot()"""
        self._phase = APIPhase.BEFORE_SHOULD_INIT

    def verify_should_init(self) -> None:
        """Check should_init()"""
        if self._phase != APIPhase.BEFORE_SHOULD_INIT:
            raise RuntimeError(
                    'Please check whether to run f_init using should_init()'
                    ' after resuming, and before trying to save a snapshot.')

    def should_init_done(self) -> None:
        """Update phase on successful should_init()"""
        self._phase = APIPhase.BEFORE_SHOULD_SAVE_SNAPSHOT

    def verify_should_save_snapshot(self) -> None:
        """Check should_save_snapshot()"""
        if self._phase != APIPhase.BEFORE_SHOULD_SAVE_SNAPSHOT:
            self._generic_error_messages('should_save_snapshot')
            raise RuntimeError()  # should be unreachable

    def should_save_snapshot_done(self, should_save: bool) -> None:
        """Update phase on successful should_save_snapshot().

        Args:
            should_save: Whether we should save or not.
        """
        if should_save:
            self._phase = APIPhase.BEFORE_SAVE_SNAPSHOT

    def verify_save_snapshot(self) -> None:
        """Check save_snapshot()"""
        if self._phase != APIPhase.BEFORE_SAVE_SNAPSHOT:
            self._generic_error_messages('save_snapshot')
            raise RuntimeError()  # should be unreachable

    def save_snapshot_done(self) -> None:
        """Update phase on successful save_snapshot()"""
        self._phase = APIPhase.BEFORE_SHOULD_SAVE_SNAPSHOT

    def verify_should_save_final_snapshot(self) -> None:
        """Check should_save_final_snapshot()."""
        if self._phase != APIPhase.BEFORE_SHOULD_SAVE_SNAPSHOT:
            self._generic_error_messages('should_save_final_snapshot')
            raise RuntimeError()  # should be unreachable

    def should_save_final_snapshot_done(self, should_save: bool) -> None:
        """Update phase on successful should_save_snapshot().

        Args:
            should_save: Whether we should save or not.
        """
        if should_save:
            self._phase = APIPhase.BEFORE_SAVE_FINAL_SNAPSHOT
        else:
            self._phase = APIPhase.BEFORE_REUSE_INSTANCE

    def verify_save_final_snapshot(self) -> None:
        """Check should_save_final_snapshot()"""
        if self._phase != APIPhase.BEFORE_SAVE_FINAL_SNAPSHOT:
            self._generic_error_messages('save_final_snapshot')
            raise RuntimeError()  # should be unreachable

    def save_final_snapshot_done(self) -> None:
        """Updates state on successful save_final_snapshot()"""
        self._phase = APIPhase.BEFORE_REUSE_INSTANCE



import multiprocessing as mp
import os
from pathlib import Path
import pytest
import subprocess
import sys

from libmuscle import Instance, Message
from ymmsl import Operator

from .conftest import skip_if_python_only


def run_macro(instance_id: str, muscle_manager: str):
    sys.argv.append('--muscle-instance={}'.format(instance_id))
    sys.argv.append('--muscle-manager={}'.format(muscle_manager))
    macro()


def macro():
    instance = Instance({
            Operator.O_I: ['out'],
            Operator.S: ['in']})

    while instance.reuse_instance():
        # f_init
        assert instance.get_setting('test1') == 13

        for i in range(2):
            # o_i
            instance.send('out', Message(i * 10.0, (i + 1) * 10.0, 'testing'))

            # s/b
            msg = instance.receive('in')
            assert msg.data == 'testing back {}'.format(i)
            assert msg.timestamp == i * 10.0


@skip_if_python_only
def test_mpi_macro_micro(
        tmpdir, mmp_server_process_simple, mpirun_outfile_arg):
    # only run this if MPI is enabled
    if 'MUSCLE_ENABLE_CPP_MPI' not in os.environ:
        pytest.skip('MPI support was not detected')

    # create C++ micro model
    # see libmuscle/cpp/src/libmuscle/tests/micro_model_test.cpp
    cpp_build_dir = Path(__file__).parents[1] / 'libmuscle' / 'cpp' / 'build'
    env = os.environ.copy()
    lib_paths = [cpp_build_dir / 'msgpack' / 'msgpack' / 'lib']
    if 'LD_LIBRARY_PATH' in env:
        env['LD_LIBRARY_PATH'] += ':' + ':'.join(map(str, lib_paths))
    else:
        env['LD_LIBRARY_PATH'] = ':'.join(map(str, lib_paths))

    env['MUSCLE_MANAGER'] = mmp_server_process_simple

    cpp_test_dir = cpp_build_dir / 'libmuscle' / 'tests'
    mpi_test_micro = cpp_test_dir / 'mpi_micro_model_test'
    out_file = tmpdir + '/mpi_micro.log'
    micro_result = subprocess.Popen(
            ['mpirun', '-np', '2', mpirun_outfile_arg, out_file,
             str(mpi_test_micro), '--muscle-instance=micro'], env=env)

    # run macro model
    macro_process = mp.Process(target=run_macro,
                               args=('macro', mmp_server_process_simple))
    macro_process.start()

    # check results
    micro_result.wait()
    assert micro_result.returncode == 0
    macro_process.join()
    assert macro_process.exitcode == 0



import pytest
from ymmsl import Reference, Settings

from libmuscle.settings_manager import SettingsManager


@pytest.fixture
def settings_manager() -> SettingsManager:
    return SettingsManager()


def test_create(settings_manager):
    assert len(settings_manager.base) == 0
    assert len(settings_manager.overlay) == 0


def test_get_setting(settings_manager):
    ref = Reference
    settings_manager.base[ref('test')] = 13
    assert settings_manager.get_setting(ref('instance'), ref('test')) == 13

    settings_manager.overlay[ref('test2')] = 14
    assert settings_manager.get_setting(ref('instance'), ref('test2')) == 14

    settings_manager.base[ref('test2')] = 'test'
    assert settings_manager.get_setting(ref('instance'), ref('test2')) == 14

    settings_manager.overlay = Settings()
    assert settings_manager.get_setting(ref('instance'), ref('test2')) == \
        'test'

    settings_manager.base[ref('test3')] = 'base_test3'
    settings_manager.base[ref('instance.test3')] = 'base_instance_test3'
    assert settings_manager.get_setting(ref('instance'), ref('test3')) == \
        'base_instance_test3'
    assert settings_manager.get_setting(ref('instance2'), ref('test3')) == \
        'base_test3'

    settings_manager.overlay[ref('test3')] = 'overlay_test3'
    settings_manager.overlay[ref('instance.test3')] = 'overlay_instance_test3'
    assert settings_manager.get_setting(ref('instance'), ref('test3')) == \
        'overlay_instance_test3'
    assert settings_manager.get_setting(ref('instance2'), ref('test3')) == \
        'overlay_test3'

    settings_manager.base[ref('instance.test4')] = 'base_test4'
    settings_manager.overlay[ref('test4')] = 'overlay_test4'
    assert settings_manager.get_setting(ref('instance'), ref('test4')) == \
        'base_test4'

    assert settings_manager.get_setting(ref('instance[10]'), ref('test4')
                                        ) == 'base_test4'

    settings_manager.base[ref('instance[10].test5')] = 'base_test5'
    settings_manager.overlay[ref('test5')] = 'overlay_test5'
    assert settings_manager.get_setting(ref('instance'), ref('test5')) == \
        'overlay_test5'
    assert settings_manager.get_setting(ref('instance[10]'), ref('test5')
                                        ) == 'base_test5'
    assert settings_manager.get_setting(ref('instance[11]'), ref('test5')
                                        ) == 'overlay_test5'



class TransportClient:
    """A client that connects to an MCP server.

    This is a base class for MCP Transport Clients. An MCP Transport
    Client connects to an MCP Transport Server over some communication
    protocol, requests messages from it, and returns responses.
    """
    @staticmethod
    def can_connect_to(location: str) -> bool:
        """Whether this client class can connect to the given location.

        Args:
            location: The location to potentially connect to.

        Returns:
            True iff this class can connect to this location.
        """
        raise NotImplementedError()     # pragma: no cover

    def call(self, request: bytes) -> bytes:
        """Send a request to the server and receive the response.

        This is a blocking call.

        Args:
            request: The request to send

        Returns:
            The received response
        """
        raise NotImplementedError()     # pragma: no cover

    def close(self) -> None:
        """Closes this client.

        This closes any connections this client has and/or performs
        other shutdown activities as needed.
        """
        raise NotImplementedError()     # pragma: no cover



from typing import List, Optional

import numpy as np


class Grid:
    """Represents a grid of data to send or receive.

    Note that for received grids, the array of data is a read-only NumPy array.
    If you have another array that you want to put the received data into, use
    :external:py:func:`np.copyto(dest, source) <numpy.copyto>` to copy the
    contents of the received array across into your destination array. If you
    don't have an array yet and want a writable version of the received array,
    use :external:py:meth:`array.copy()<numpy.ndarray.copy>` to create a
    writable copy. See the tutorial for examples.

    Attributes:
        array (np.ndarray): An array of data
        indexes (Optional[List[str]]): The names of the array's indexes.
    """
    def __init__(
            self, array: np.ndarray, indexes: Optional[List[str]] = None
            ) -> None:
        """Creates a Grid object.

        A Grid object represents an multi-dimensional array of data. It
        has a type, a shape, and optionally a list of index names.

        Supported data types are 4- and 8-byte integers
        (:external:py:attr:`numpy.int32`, :external:py:attr:`numpy.int64`),
        4- and 8-byte floats (:external:py:attr:`numpy.float32`,
        :external:py:attr:`numpy.float64`), and booleans
        (:external:py:class:`numpy.bool_`, :external:py:attr:`numpy.bool8`). The
        ``data`` argument must be a NumPy array of one of those types.

        If ``indexes`` is given, then it must be a list of strings of
        the same length as the number of dimensions of ``data``, and
        contain the names of the indexes of the array. For a 2D
        Cartesian grid, these may be ``'x'`` and ``'y'`` for example,
        or for a polar grid, ``'phi'`` and ``'rho'``.

        Args:
            array: An array of data, of a supported type (see above).
            indexes: Names of the indexes (see above).
        """
        if indexes is not None:
            if len(indexes) != array.ndim:
                raise ValueError(
                        'Number of indexes must match number of array'
                        ' dimensions')

        self.array = array
        self.indexes = indexes



import enum
import logging
import multiprocessing as mp
from pathlib import Path
from typing import Optional

from ymmsl import Implementation, Reference, ResourceRequirements

from libmuscle.planner.planner import Resources


class ProcessStatus(enum.Enum):
    """Status of a process (instance)."""
    STARTED = 0
    RUNNING = 1
    SUCCESS = 2
    ERROR = 3
    CANCELED = 4

    def is_finished(self) -> bool:
        """Returns whether the job is finished.

        Canceled jobs are considered finished, even if they've never
        run.
        """
        return self in (
                ProcessStatus.SUCCESS, ProcessStatus.ERROR,
                ProcessStatus.CANCELED)


class Process:
    """Represents a process.

    Attributes:
        instance: Name of instance this is the process of
        resources: The resources allocated to this process
        status: Current status of the process
        exit_code: Exit code, if status is ERROR
        error_msg: Error message, if status is ERROR
    """
    def __init__(self, instance: Reference, resources: Resources) -> None:
        """Create a Process object.

        Args:
            resources: Resources allocated to this process.
        """
        self.instance = instance
        self.resources = resources
        self.status = ProcessStatus.STARTED
        self.exit_code = None   # type: Optional[int]
        self.error_msg = None   # type: Optional[str]


class InstantiatorRequest:
    """Base class for requests to an instantiator."""
    pass


class ShutdownRequest(InstantiatorRequest):
    """Requests shutting down the background process.

    The process will stop once all running processes are done.
    """
    pass


class InstantiationRequest(InstantiatorRequest):
    """Requests instantiating a new process.

    Attributes:
        instance: The name of the instance
        implementation: The implementation to start for it
        resources: The resources to start it on
    """
    def __init__(
            self, instance: Reference, implementation: Implementation,
            res_req: ResourceRequirements, resources: Resources, instance_dir:
            Path, work_dir: Path, stdout_path: Path, stderr_path: Path
            ) -> None:
        """Create an InstantiationRequest.

        Args:
            instance: The name of the instance
            implementation: The implementation to start for it
            res_req: The resource requirements for this instance
            resources: The resources to instantiate on
            instance_dir: The main directory for this instance
            work_dir: The directory in which to start it
            stdout_path: Path of file to redirect stdout to
            stderr_path: Path of file to redirect stderr to
        """
        self.instance = instance
        self.implementation = implementation
        self.res_req = res_req
        self.resources = resources
        self.instance_dir = instance_dir
        self.work_dir = work_dir
        self.stdout_path = stdout_path
        self.stderr_path = stderr_path


class CancelAllRequest(InstantiatorRequest):
    """Requests stopping all running processes."""
    pass


class CrashedResult:
    """Signals that the instantiator process crashed."""
    pass


class QueueingLogHandler(logging.Handler):
    """A logging Handler that enqueues records."""
    def __init__(self, queue: mp.Queue) -> None:
        """Create a QueueingLogHandler.

        Args:
            level: The level of this handler
        """
        super().__init__()
        self._queue = queue

    def emit(self, record: logging.LogRecord) -> None:
        """Emit the record by enqueueing it.

        Args:
            record: A log record to enqueue.
        """
        self._queue.put(record)



import pytest
from unittest.mock import patch

from ymmsl import Settings

from libmuscle.api_guard import APIGuard
from libmuscle.communicator import Message
from libmuscle.mmp_client import MMPClient


@pytest.fixture
def mocked_mmp_client():
    with patch('libmuscle.mmp_client.TcpTransportClient') as mock_ttc:
        yield MMPClient(''), mock_ttc.return_value


@pytest.fixture
def message() -> Message:
    return Message(0.0, None, b'test', Settings())


@pytest.fixture
def message2() -> Message:
    return Message(0.0, None, {'test': 17}, Settings())


@pytest.fixture
def guard() -> APIGuard:
    return APIGuard(True)
